/*****************************************************************/
/*  ELEC 999                                                     */
/*  University of Nebraska-Lincoln                               */
/*  Department of Electrical Engineering                         */
/*  David J. Russell                                             */
/*  5-26-08                                                      */
/*                                                               */
/* 01/30/09 - Added % distribution to output type, version 1.1   */
/*****************************************************************/

#define _CNFADT

/*********************************************************************************/
/* Included Header Files */
/*********************************************************************************/
#include "Globals.h"
#include "VectorADT.h"
#include "CNFGrammarADT.h"
#include "Parameters.h"

/*********************************************************************************/
/* Private Module Constants */
/*********************************************************************************/

static const UBYTE m_DNAInverseLookup[ALPHABETIC_CHARACTERS] = 
{
   'T', 'X', 'G', 'X', 'X', 'X', 'C', 'X', 'X', 'X', 'X', 'X', 'X', 
   'X', 'X', 'X', 'X', 'X', 'X', 'A', 'X', 'X', 'X', 'X', 'X', 'X', 'X'
};

static const UBYTE m_RNAInverseLookup[ALPHABETIC_CHARACTERS] = 
{
   'U', 'X', 'G', 'X', 'X', 'X', 'C', 'X', 'X', 'X', 'X', 'X', 'X', 
   'X', 'X', 'X', 'X', 'X', 'X', 'X', 'A', 'X', 'X', 'X', 'X', 'X', 'X'
};

/*********************************************************************************/
/* Private Type Definitions */
/*********************************************************************************/

/*********************************************************************************/
/* Private Module Function Prototypes */
/*********************************************************************************/
static void CNFADT_Free (CNFADT_NODE_TYPE **Location);
static void CNFADT_DeleteRuleFromDictionary(CNFADT_DICTIONARY_TYPE *Dictionary, 
                                             CNFADT_NODE_TYPE **HeadNode);
static BOOLEAN CNFADT_RuleType2BodiesAreEqual(CNFADT_BODY_TYPE BodyA,
                                                CNFADT_BODY_TYPE BodyB);
static BOOLEAN CNFADT_RuleType2BodiesAreReverseComplements(CNFADT_BODY_TYPE BodyA,
                                                            CNFADT_BODY_TYPE BodyB);
static BOOLEAN CNFADT_RuleType2BodyIsReverseComplementWithItself(CNFADT_RULE_ENTRY Rule);

/*********************************************************************************/
/* Private Module Variables */
/*********************************************************************************/

/*********************************************************************************/
/* Public and Private Module Functions */
/*********************************************************************************/

void CNFADT_CreateDictionary(CNFADT_DICTIONARY_TYPE *Dictionary)
{
   Dictionary->List = NULL;
   Dictionary->CurrentHeadLabel = 0;
}

/*********************************************************************************/

void CNFADT_DeleteDictionary(CNFADT_DICTIONARY_TYPE *Dictionary)
{
   CNFADT_NODE_TYPE *TempPointer;
   
   while (Dictionary->List != NULL)
   {
      TempPointer = Dictionary->List;
      Dictionary->List = Dictionary->List->Next;

      CNFADT_Free(&TempPointer);
   }
}

/*********************************************************************************/

void CNFADT_Free (CNFADT_NODE_TYPE **Location)
{
   if (*Location != NULL)
   {
      free(*Location);
      *Location = NULL;
   }
}

/*********************************************************************************/

void CNFADT_AddRuleToDictionary(CNFADT_DICTIONARY_TYPE *Dictionary, 
                                 CNFADT_RULE_AND_METRIC_TYPE NewEntry,
                                 CNFADT_NODE_TYPE **NewNodePointer,
                                 BOOLEAN *ReverseComplement,
                                 BOOLEAN PresetHeadCountToZero)
{
   CNFADT_NODE_TYPE *Location;
   CNFADT_NODE_TYPE *PreviousLocation;
   CNFADT_NODE_TYPE *NewNode;
   BOOLEAN NeedNewRule = TRUE;

   if (NewEntry.Rule.Type == CNFADT_RULE_TYPE1)
   {
      if (CNFADT_FindRuleType1Body(Dictionary, 
                                    NewEntry.Rule.Body, 
                                    ReverseComplement,
                                    &Location,
                                    &PreviousLocation) == TRUE)
      {
         NeedNewRule = FALSE;
      }
   }
   else
   {
      if (CNFADT_FindRuleType2Body(Dictionary, 
                                    NewEntry.Rule.Body, 
                                    ReverseComplement,
                                    &Location,
                                    &PreviousLocation) == TRUE)
      {
         NeedNewRule = FALSE;
      }
   }

   if (NeedNewRule == TRUE)
   {
      NewNode = (CNFADT_NODE_TYPE *) malloc (sizeof(CNFADT_NODE_TYPE));

      if (NewNode == NULL)
      {
         printf("Error: malloc failed in CNFADT_AddRuleToDictionary.\n");
         exit(1);
      }
      else
      {
         NewNode->Entry = NewEntry;
         NewNode->Entry.Rule.Head = Dictionary->CurrentHeadLabel;
         Dictionary->CurrentHeadLabel += 1;
         if (PresetHeadCountToZero == TRUE)
         {
            NewNode->Entry.Metric.HeadRepeatCount = 0;
         }
         else
         {
            NewNode->Entry.Metric.HeadRepeatCount = 1;
         }
         NewNode->Entry.Metric.SelfReverseComplement = CNFADT_RuleType2BodyIsReverseComplementWithItself(NewEntry.Rule);
         NewNode->Entry.Metric.InABodyLinkCount = 0;
         if (NewEntry.Rule.Type == CNFADT_RULE_TYPE2)
         {
            /*
               These counts are needed for when we go to delete a grammar rule, we
               need to make sure we don't have some rule above us, pointing to us.
            */
            NewNode->Entry.Rule.Body.Type2Body[0].Entry->Metric.InABodyLinkCount += 1;
            NewNode->Entry.Rule.Body.Type2Body[1].Entry->Metric.InABodyLinkCount += 1;
         }
         NewNode->Next = NULL;         
         NewNode->Prev = NULL;         
      }

      *NewNodePointer = NewNode;

      if (PreviousLocation == NULL)
      {
         NewNode->Next = Dictionary->List;
         Dictionary->List = NewNode;
      }
      else
      {
         NewNode->Next = PreviousLocation->Next;
         PreviousLocation->Next = NewNode;
      }
      if (Location != NULL)
      {
         NewNode->Next->Prev = NewNode;
      }
      NewNode->Prev = PreviousLocation;
   }
   else
   {
      *NewNodePointer = Location;
      Location->Entry.Metric.HeadRepeatCount += 1;
   }
}

/*********************************************************************************/

void CNFADT_DeleteRuleFromDictionary(CNFADT_DICTIONARY_TYPE *Dictionary, 
                                       CNFADT_NODE_TYPE **HeadNode)
{
   CNFADT_NODE_TYPE *TempPointer;

   if ((*HeadNode)->Entry.Metric.InABodyLinkCount > 0)
   {
      printf("Error: deleting a rule that is in the body of some other production.\n");
      exit(1);
   }
   else
   {
      if ((*HeadNode)->Entry.Rule.Type == CNFADT_RULE_TYPE2)
      {
         /*
            These counts are needed for when we go to delete a grammar rule, we
            need to make sure we don't have some rule above us, pointing to us.
         */
         if ((*HeadNode)->Entry.Rule.Body.Type2Body[0].Entry->Metric.InABodyLinkCount > 0)
         {
            (*HeadNode)->Entry.Rule.Body.Type2Body[0].Entry->Metric.InABodyLinkCount -= 1;
         }
         else
         {
            printf("Error: grammar rule body count is already 0.\n");
            exit(1);
         }

         if ((*HeadNode)->Entry.Rule.Body.Type2Body[1].Entry->Metric.InABodyLinkCount > 0)
         {
            (*HeadNode)->Entry.Rule.Body.Type2Body[1].Entry->Metric.InABodyLinkCount -= 1;
         }
         else
         {
            printf("Error: grammar rule body count is already 0.\n");
            exit(1);
         }
      }

      /*
         We will never be deleting the first node (it is a type-1 rule), nor the
         last node (it is the root of the tree).
      */
      TempPointer = (*HeadNode)->Next;
      
      (*HeadNode)->Prev->Next = (*HeadNode)->Next;
      (*HeadNode)->Next->Prev = (*HeadNode)->Prev;

      CNFADT_Free(HeadNode);      
   }
}

/*********************************************************************************/

void CNFADT_DecrementRuleCountInDictionary(CNFADT_DICTIONARY_TYPE *Dictionary, 
                                             CNFADT_NODE_TYPE **HeadNode)
{
   if ((*HeadNode)->Entry.Metric.HeadRepeatCount > 0)
   {
      (*HeadNode)->Entry.Metric.HeadRepeatCount -= 1;
   }

   if ((*HeadNode)->Entry.Metric.HeadRepeatCount == 0)
   {
      CNFADT_DeleteRuleFromDictionary(Dictionary, HeadNode);
   }
}

/*********************************************************************************/

void CNFADT_RelabelRulesInDictionary(CNFADT_DICTIONARY_TYPE *Dictionary)
{
   CNFADT_NODE_TYPE *TempPointer;
   ULONG Label;

   TempPointer = Dictionary->List;
   Label = 0;
      
   while (TempPointer != NULL)
   {
      TempPointer->Entry.Rule.Head = Label;
      TempPointer = TempPointer->Next;
      Label += 1;
   }
   Dictionary->CurrentHeadLabel = Label;
}

/*********************************************************************************/

BOOLEAN CNFADT_FindRule(CNFADT_DICTIONARY_TYPE *Dictionary, 
                        CNFADT_HEAD_TYPE HeadValue, 
                        CNFADT_NODE_TYPE **Location,
                        CNFADT_NODE_TYPE **PreviousLocation)
{
   BOOLEAN MoreToSearch = TRUE;
   BOOLEAN Found = FALSE;

   *Location = Dictionary->List;
   *PreviousLocation = NULL;
   
   while ((MoreToSearch == TRUE) && (*Location != NULL))
   {
      if ((*Location)->Entry.Rule.Head < HeadValue)
      {
         *PreviousLocation = *Location;
         *Location = (*Location)->Next;
      }
      else
      {
         MoreToSearch = FALSE;
      }
   }
   
   if (*Location != NULL)
   {
      if ((*Location)->Entry.Rule.Head == HeadValue)
      {
         Found = TRUE;
      }
   }

   return (Found);
}

/*********************************************************************************/

BOOLEAN CNFADT_FindRuleType1Body(CNFADT_DICTIONARY_TYPE *Dictionary, 
                                 CNFADT_BODY_TYPE BodyValue, 
                                 BOOLEAN *ReverseComplement,
                                 CNFADT_NODE_TYPE **Location,
                                 CNFADT_NODE_TYPE **PreviousLocation)
{
   BOOLEAN MoreToSearch = TRUE;
   BOOLEAN Found = FALSE;
   
   *Location = Dictionary->List;
   *PreviousLocation = NULL;
   *ReverseComplement = FALSE;

   while ((MoreToSearch == TRUE) && (*Location != NULL))
   {
      if ((*Location)->Entry.Rule.Type == CNFADT_RULE_TYPE1)
      {
         if ((*Location)->Entry.Rule.Body.Type1Body.Label != BodyValue.Type1Body.Label)
         {
            MoreToSearch = FALSE;
            if (PARMS_GetDisableReverseComplements() == FALSE)
            {
               switch (PARMS_GetSequenceType())
               {
                  case PARMS_SEQUENCE_DNA:
                     if ((*Location)->Entry.Rule.Body.Type1Body.Label != m_DNAInverseLookup[ALPHA_INDEX(BodyValue.Type1Body.Label)])
                     {
                        MoreToSearch = TRUE;
                     }
                     break;
                     
                  case PARMS_SEQUENCE_RNA:
                     if ((*Location)->Entry.Rule.Body.Type1Body.Label != m_RNAInverseLookup[ALPHA_INDEX(BodyValue.Type1Body.Label)])
                     {
                        MoreToSearch = TRUE;
                     }
                     break;

                  case PARMS_SEQUENCE_AMINO_ACID:
                     MoreToSearch = TRUE;
                     break;
               }
            }
            else
            {
               MoreToSearch = TRUE;
            }
         }
         else
         {
            MoreToSearch = FALSE;
         }
      }

      if (MoreToSearch == TRUE)
      {
         *PreviousLocation = *Location;
         *Location = (*Location)->Next;
      }
   }
   
   if (*Location != NULL)
   {
      if ((*Location)->Entry.Rule.Type == CNFADT_RULE_TYPE1)
      {
         if ((*Location)->Entry.Rule.Body.Type1Body.Label == BodyValue.Type1Body.Label)
         {
            Found = TRUE;
         }
         else
         {
            if (PARMS_GetDisableReverseComplements() == FALSE)
            {
               switch (PARMS_GetSequenceType())
               {
                  case PARMS_SEQUENCE_DNA:
                     if ((*Location)->Entry.Rule.Body.Type1Body.Label == m_DNAInverseLookup[ALPHA_INDEX(BodyValue.Type1Body.Label)])
                     {
                        Found = TRUE;
                        *ReverseComplement = TRUE;
                     }
                     break;
                     
                  case PARMS_SEQUENCE_RNA:
                     if ((*Location)->Entry.Rule.Body.Type1Body.Label == m_RNAInverseLookup[ALPHA_INDEX(BodyValue.Type1Body.Label)])
                     {
                        Found = TRUE;
                        *ReverseComplement = TRUE;
                     }
                     break;

                  case PARMS_SEQUENCE_AMINO_ACID:
                     break;
               }
            }
         }
      }
   }

   return (Found);
}

/*********************************************************************************/

BOOLEAN CNFADT_FindRuleType2Body(CNFADT_DICTIONARY_TYPE *Dictionary, 
                                 CNFADT_BODY_TYPE BodyValue, 
                                 BOOLEAN *ReverseComplement,
                                 CNFADT_NODE_TYPE **Location,
                                 CNFADT_NODE_TYPE **PreviousLocation)
{
   BOOLEAN MoreToSearch = TRUE;
   BOOLEAN Found = FALSE;

   *Location = Dictionary->List;
   *PreviousLocation = NULL;
   *ReverseComplement = FALSE;
   
   while ((MoreToSearch == TRUE) && (*Location != NULL))
   {
      if ((*Location)->Entry.Rule.Type == CNFADT_RULE_TYPE2)
      {
         if (CNFADT_RuleType2BodiesAreEqual((*Location)->Entry.Rule.Body, BodyValue) == FALSE)
         {
            MoreToSearch = FALSE;
            if (PARMS_GetDisableReverseComplements() == FALSE)
            {
               switch (PARMS_GetSequenceType())
               {
                  case PARMS_SEQUENCE_DNA:
                  case PARMS_SEQUENCE_RNA:
                     if (CNFADT_RuleType2BodiesAreReverseComplements((*Location)->Entry.Rule.Body, BodyValue) == FALSE)
                     {
                        MoreToSearch = TRUE;
                     }
                     break;
                     
                  case PARMS_SEQUENCE_AMINO_ACID:
                     MoreToSearch = TRUE;
                     break;
               }
            }
            else
            {
               MoreToSearch = TRUE;
            }
         }
         else
         {
            MoreToSearch = FALSE;
         }
      }

      if (MoreToSearch == TRUE)
      {
         *PreviousLocation = *Location;
         *Location = (*Location)->Next;
      }
   }
   
   if (*Location != NULL)
   {
      if ((*Location)->Entry.Rule.Type == CNFADT_RULE_TYPE2)
      {
         if (CNFADT_RuleType2BodiesAreEqual((*Location)->Entry.Rule.Body, BodyValue) == TRUE)
         {
            Found = TRUE;
         }
         else
         {
            if (PARMS_GetDisableReverseComplements() == FALSE)
            {
               switch (PARMS_GetSequenceType())
               {
                  case PARMS_SEQUENCE_DNA:
                  case PARMS_SEQUENCE_RNA:
                     if (CNFADT_RuleType2BodiesAreReverseComplements((*Location)->Entry.Rule.Body, BodyValue) == TRUE)
                     {
                        Found = TRUE;
                        *ReverseComplement = TRUE;
                     }
                     break;
                     
                  case PARMS_SEQUENCE_AMINO_ACID:
                     break;
               }
            }
         }
      }
   }

   return (Found);
}

/*********************************************************************************/

BOOLEAN CNFADT_RuleType2BodiesAreEqual(CNFADT_BODY_TYPE BodyA,
                                       CNFADT_BODY_TYPE BodyB)
{
   BOOLEAN BodiesEqual = FALSE;

   if ((BodyA.Type2Body[0].Entry->Rule.Head == BodyB.Type2Body[0].Entry->Rule.Head) &&
       (BodyA.Type2Body[1].Entry->Rule.Head == BodyB.Type2Body[1].Entry->Rule.Head))
   {
      if ((BodyA.Type2Body[0].Entry->Metric.SelfReverseComplement != BodyB.Type2Body[0].Entry->Metric.SelfReverseComplement) ||
          (BodyA.Type2Body[1].Entry->Metric.SelfReverseComplement != BodyB.Type2Body[1].Entry->Metric.SelfReverseComplement))
      {
         printf("Error: the same grammar rules have different information.\n");
         exit(1);
      }
      else
      {
         if (((BodyA.Type2Body[0].ReverseComplement == BodyB.Type2Body[0].ReverseComplement) ||
              (BodyA.Type2Body[0].Entry->Metric.SelfReverseComplement == TRUE)) &&
             ((BodyA.Type2Body[1].ReverseComplement == BodyB.Type2Body[1].ReverseComplement) ||
              (BodyA.Type2Body[1].Entry->Metric.SelfReverseComplement == TRUE)))
         {
            BodiesEqual = TRUE;
         }
      }
   }

   return (BodiesEqual);
}

/*********************************************************************************/

BOOLEAN CNFADT_RuleType2BodiesAreReverseComplements(CNFADT_BODY_TYPE BodyA,
                                                      CNFADT_BODY_TYPE BodyB)
{
   BOOLEAN BodiesComplement = FALSE;
   
   if ((BodyA.Type2Body[0].Entry->Rule.Head == BodyB.Type2Body[1].Entry->Rule.Head) &&
       (BodyA.Type2Body[1].Entry->Rule.Head == BodyB.Type2Body[0].Entry->Rule.Head))
   {
      if ((BodyA.Type2Body[0].Entry->Metric.SelfReverseComplement != BodyB.Type2Body[1].Entry->Metric.SelfReverseComplement) ||
          (BodyA.Type2Body[1].Entry->Metric.SelfReverseComplement != BodyB.Type2Body[0].Entry->Metric.SelfReverseComplement))
      {
         printf("Error: the same grammar rules have different information.\n");
         exit(1);
      }
      else
      {
         if (((BodyA.Type2Body[0].ReverseComplement == NOT(BodyB.Type2Body[1].ReverseComplement)) ||
              (BodyA.Type2Body[0].Entry->Metric.SelfReverseComplement == TRUE)) &&
             ((BodyA.Type2Body[1].ReverseComplement == NOT(BodyB.Type2Body[0].ReverseComplement)) ||
              (BodyA.Type2Body[1].Entry->Metric.SelfReverseComplement == TRUE)))
         {
            BodiesComplement = TRUE;
         }
      }
   }
   
   return (BodiesComplement);
}

/*********************************************************************************/

BOOLEAN CNFADT_RuleType2BodyIsReverseComplementWithItself(CNFADT_RULE_ENTRY Rule)
{
   BOOLEAN BodyComplements = FALSE;

   if (Rule.Type == CNFADT_RULE_TYPE2)
   {
      if (Rule.Body.Type2Body[0].Entry->Rule.Head == Rule.Body.Type2Body[1].Entry->Rule.Head)
      {
         if (Rule.Body.Type2Body[0].Entry->Metric.SelfReverseComplement != Rule.Body.Type2Body[1].Entry->Metric.SelfReverseComplement)
         {
            printf("Error: the same grammar rules have different information.\n");
            exit(1);
         }
         else
         {
            if (Rule.Body.Type2Body[0].Entry->Metric.SelfReverseComplement == TRUE)
            {
               BodyComplements = TRUE;
            }
            else
            {
               if (Rule.Body.Type2Body[0].ReverseComplement == NOT(Rule.Body.Type2Body[1].ReverseComplement))
               {
                  BodyComplements = TRUE;
               }
            }
         }
      }
   }
      
   return (BodyComplements);
}
