/*****************************************************************/
/*  ELEC 999                                                     */
/*  University of Nebraska-Lincoln                               */
/*  Department of Electrical Engineering                         */
/*  David J. Russell                                             */
/*  5-26-08                                                      */
/*                                                               */
/* 01/30/09 - Added % distribution to output type, version 1.1   */
/*****************************************************************/

#ifdef _CNFADT
#define CNFADT_EXTERN
#else
#define CNFADT_EXTERN extern
#endif

/******************************************************************/
/* Public Module Constants                                        */
/******************************************************************/
#define CNFADT_TYPE2_RULE_NUMBER_OF_ELEMENTS 2

/******************************************************************/
/* Public Type Definitions                                        */
/******************************************************************/
typedef enum
{
   CNFADT_RULE_TYPE1,   /* A -> a  */
   CNFADT_RULE_TYPE2    /* A -> BC */
} CNFADT_RULE_TYPE;

typedef ULONG CNFADT_HEAD_TYPE;

typedef struct
{
   UBYTE Label;
} CNFADT_TYPE1_BODY_TYPE;

typedef struct
{
   BOOLEAN ReverseComplement;
   struct CNF_rule_and_metric *Entry;
} CNFADT_TYPE2_BODY_TYPE;

typedef union
{
   CNFADT_TYPE1_BODY_TYPE Type1Body;
   CNFADT_TYPE2_BODY_TYPE Type2Body[CNFADT_TYPE2_RULE_NUMBER_OF_ELEMENTS];
} CNFADT_BODY_TYPE;

typedef struct
{
   CNFADT_RULE_TYPE Type;
   CNFADT_HEAD_TYPE Head;
   CNFADT_BODY_TYPE Body;
} CNFADT_RULE_ENTRY;

typedef struct
{
   ULONG InvertedDepth;
   ULONG HeadRepeatCount;
   ULONG InABodyLinkCount;
   BOOLEAN SelfReverseComplement;

   ULONG DistanceFromStructuralAddition;
   ULONG NonOverlappingHeadRepeatCount;
   ULONG BodyDepthDifference;
   BOOLEAN BothBodiesAreStructuralPieces;
   
   VECTORADT_VECTOR_TYPE StructuralComponents;
   VECTORADT_VECTOR_TYPE ComplementComponents;
   VECTORADT_VECTOR_TYPE SelfComplementComponents;
   
   VECTORADT_VECTOR_TYPE SumOfComponents;
   VECTORADT_VECTOR_TYPE AbsDifferenceOfComponents;

   BOOLEAN ComponentVectorsSet;
} CNFADT_RULE_METRIC_ENTRY;

typedef struct CNF_rule_and_metric
{
   CNFADT_RULE_ENTRY Rule;
   CNFADT_RULE_METRIC_ENTRY Metric;
} CNFADT_RULE_AND_METRIC_TYPE;

typedef struct CNF_node
{
   CNFADT_RULE_AND_METRIC_TYPE Entry;
   struct CNF_node *Next;
   struct CNF_node *Prev;
} CNFADT_NODE_TYPE;

typedef struct
{
   ULONG CurrentHeadLabel;
   CNFADT_NODE_TYPE *List;
} CNFADT_DICTIONARY_TYPE;

/*********************************************************************************/
/* Public Module Variables */
/*********************************************************************************/

/*********************************************************************************/
/* Public Module Function Prototypes */
/*********************************************************************************/
void CNFADT_CreateDictionary(CNFADT_DICTIONARY_TYPE *Dictionary);
void CNFADT_DeleteDictionary(CNFADT_DICTIONARY_TYPE *Dictionary);
void CNFADT_AddRuleToDictionary(CNFADT_DICTIONARY_TYPE *Dictionary, 
                                 CNFADT_RULE_AND_METRIC_TYPE NewEntry,
                                 CNFADT_NODE_TYPE **NewNodePointer,
                                 BOOLEAN *ReverseComplement,
                                 BOOLEAN PresetHeadCountToZero);
void CNFADT_DecrementRuleCountInDictionary(CNFADT_DICTIONARY_TYPE *Dictionary, 
                                             CNFADT_NODE_TYPE **HeadNode);
void CNFADT_RelabelRulesInDictionary(CNFADT_DICTIONARY_TYPE *Dictionary);
BOOLEAN CNFADT_FindRule(CNFADT_DICTIONARY_TYPE *Dictionary, 
                        CNFADT_HEAD_TYPE HeadValue, 
                        CNFADT_NODE_TYPE **Location, 
                        CNFADT_NODE_TYPE **PreviousLocation);
BOOLEAN CNFADT_FindRuleType1Body(CNFADT_DICTIONARY_TYPE *Dictionary, 
                                 CNFADT_BODY_TYPE BodyValue, 
                                 BOOLEAN *ReverseComplement,
                                 CNFADT_NODE_TYPE **Location, 
                                 CNFADT_NODE_TYPE **PreviousLocation);                                 
BOOLEAN CNFADT_FindRuleType2Body(CNFADT_DICTIONARY_TYPE *Dictionary, 
                                 CNFADT_BODY_TYPE BodyValue, 
                                 BOOLEAN *ReverseComplement,
                                 CNFADT_NODE_TYPE **Location, 
                                 CNFADT_NODE_TYPE **PreviousLocation);
