/*****************************************************************/
/*  ELEC 999                                                     */
/*  University of Nebraska-Lincoln                               */
/*  Department of Electrical Engineering                         */
/*  David J. Russell                                             */
/*  5-26-08                                                      */
/*                                                               */
/* 01/30/09 - Added % distribution to output type, version 1.1   */
/* 01/31/09 - Added windowing across an input, version 1.2       */
/* 02/03/09 - Added unpaired repeats and modified paired counts, */
/*             version 1.3                                       */
/* 02/07/09 - Added extreme-value % removal for the stats output,*/
/*             version 1.4                                       */
/* 02/14/09 - Added html output type where text is colored to    */
/*             identify the structural pieces. Modified          */
/*             VectorADT to allow longer vectors, version 1.5    */
/* 02/16/09 - Added minimum ID parameter for labeling nodes on   */
/*             .eps outputs, version 1.6                         */
/* 02/17/09 - Added a parameter to control how many nodes below  */
/*             the root of a structural node are shown on the    */
/*             .eps outputs, version 1.7                         */
/*****************************************************************/

#ifdef _PARMS
#define PARMS_EXTERN
#else
#define PARMS_EXTERN extern
#endif

/*********************************************************************************/
/* Public Module Constants */
/*********************************************************************************/

/*********************************************************************************/
/* Public Type Definitions */
/*********************************************************************************/
typedef enum
{
   PARMS_SEQUENCE_DNA,
   PARMS_SEQUENCE_RNA,
   PARMS_SEQUENCE_AMINO_ACID
} PARMS_SEQUENCE_TYPE;

typedef enum
{
   PARMS_OUTPUT_VERBOSE_GRAMMAR_ONLY,
   PARMS_OUTPUT_COMPRESSED_GRAMMAR_ONLY,
   PARMS_OUTPUT_VERBOSE_GRAMMAR_AND_TREE,
   PARMS_OUTPUT_COMPRESSED_GRAMMAR_AND_TREE,
   PARMS_OUTPUT_STRUCTURAL_DISTRIBUTION_PERCENTAGE,
   PARMS_OUTPUT_HTML_COLORED_SEQUENCE
} PARMS_OUTPUT_TYPE;

/*********************************************************************************/
/* Public Module Variables */
/*********************************************************************************/

/*********************************************************************************/
/* Public Module Function Prototypes */
/*********************************************************************************/
BYTE *PARMS_GetInputFileName(void);
BYTE *PARMS_GetPostScriptOutputFileName(void);
BYTE *PARMS_GetGrammarOutputFileName(void);
BYTE *PARMS_GetDistributionOutputFileName(void);
BYTE *PARMS_GetHTMLOutputFileName(void);
PARMS_OUTPUT_TYPE PARMS_GetOutputType(void);
PARMS_SEQUENCE_TYPE PARMS_GetSequenceType(void);
void PARMS_SetSequenceType(PARMS_SEQUENCE_TYPE SequenceType);
BOOLEAN PARMS_GetDisableReverseComplements(void);
ULONG PARMS_GetPairedDistanceThreshold(void);
ULONG PARMS_GetUnpairedDistanceThreshold(void);
ULONG PARMS_GetWindowLength(void);
BOOLEAN PARMS_GetDisableWindow(void);
ULONG PARMS_GetMinimumStructuralLength(void);
UBYTE PARMS_GetExtremeValueRemovalPercentage(void);
ULONG PARMS_GetMinimumIDForLabel(void);
BOOLEAN PARMS_GetEnableSelectiveNodeOutput(void);
ULONG PARMS_GetSelectiveNodeOutputPathDistance(void);
BOOLEAN PARMS_ProcessCommandLine(INT argc, BYTE **argv);
void PARMS_Printf(BYTE *String, BOOLEAN Show);
void PARMS_PrintfProgress(ULONG Current, ULONG Total, BOOLEAN Show);
