/*****************************************************************/
/*  ELEC 999                                                     */
/*  University of Nebraska-Lincoln                               */
/*  Department of Electrical Engineering                         */
/*  David J. Russell                                             */
/*  5-17-08                                                      */
/*****************************************************************/

#define _PS

/*
   All code in this module was taken and modified from the Cgraph 
   PostScript plotting library in C [version 2.04, 1998-08-29].

   For the current version, see the original source:
   http://totoro.berkeley.edu/software/A_Cgraph.html

   ftp://pinoko.berkeley.edu/pub/software/Cgraph2.04.tar.gz
   (version number 2.04 may change)

   Izumi Ohzawa, Ph.D.
   izumi@pinoko.berkeley.edu
*/

/*********************************************************************************/
/* Included Header Files */
/*********************************************************************************/
#include "Globals.h"
#include "PostScript.h"
#include "Parameters.h"

/*********************************************************************************/
/* Private Module Constants */
/*********************************************************************************/

#define PS_MAX_STRING_LENGTH                          128

#define PS_NUMBER_PSFONTS                             35

#define PS_MINBIT                                     0x0001
#define PS_MAXBIT                                     0x0400

/*********************************************************************************/
/* Private Type Definitions */
/*********************************************************************************/
typedef enum
{
   PS_ROUND_DOWN = 0,
   PS_ROUND_UP,
   PS_ROUND_EVEN
} PS_ROUND_TYPE;

/*********************************************************************************/
/* Private Module Function Prototypes */
/*********************************************************************************/
static void PS_LoadMark(PS_MARKER_TYPE Type);
static void PS_XMesh (void);
static void PS_YMesh (void);
static void PS_LogXMesh (void);
static void PS_XMeshLineSet (FLOAT Min);
static void PS_XMinMeshLineSet (FLOAT Min);
static void PS_LogYMesh (void);
static void PS_YMeshLineSet (FLOAT Min);
static void PS_YMinMeshLineSet (FLOAT Min);
static void PS_FixString (BYTE *Text, BYTE *FixedText, BOOLEAN FixMinus);
static UBYTE PS_FindSignificantDecimalPoint (FLOAT Min, FLOAT Range, UBYTE NumberSeparation, FLOAT TickSeparation);
static void PS_FormAxisNumber (FLOAT Value, UBYTE SignificantDecimal, BYTE *Line);
static void PS_CleanExponentialString (BYTE *Line, UBYTE ExponentIndex);
static FLOAT PS_Trunc(FLOAT Value, PS_ROUND_TYPE Round);
static void PS_XLogTicNumbering (FLOAT Offset, 
                                 SHORT TextTop, 
                                 SHORT TextBottom, 
                                 SHORT NumberDistanceFromTic, 
                                 SHORT TextTop10, 
                                 SHORT TextBottom10,
                                 DOUBLE XScale);
static void PS_XSetTics (FLOAT Min, FLOAT Top, FLOAT Bottom);
static void PS_AllXSetTics (FLOAT Top, FLOAT Bottom);
static void PS_XSetNumbers (FLOAT Start, 
                              SHORT NumberDistanceFromTic, 
                              FLOAT Offset, 
                              SHORT TextTop, 
                              SHORT TextBottom,
                              DOUBLE XScale);
static void PS_XSetMinNumbers (FLOAT Start, 
                              SHORT NumberDistanceFromTic, 
                              FLOAT Offset, 
                              SHORT TextTop, 
                              SHORT TextBottom,
                              DOUBLE XScale);
static FLOAT PS_XBottom (FLOAT Multiplier);
static FLOAT PS_YBottom (FLOAT Multiplier);
static void PS_YLogTicNumbering (FLOAT Offset, 
                                 SHORT TextRight, 
                                 SHORT TextLeft, 
                                 SHORT NumberDistanceFromTic, 
                                 SHORT TextRight10, 
                                 SHORT TextLeft10,
                                 DOUBLE YScale);
static void PS_YSetTics (FLOAT Min, FLOAT Right, FLOAT Left);
static void PS_YSetMinTics (FLOAT Min, FLOAT Right, FLOAT Left);
static void PS_AllYSetTics (FLOAT Left, FLOAT Right);
static void PS_YSetNumbers (FLOAT Start, 
                              SHORT NumberDistanceFromTic, 
                              FLOAT Offset, 
                              SHORT TextRight, 
                              SHORT TextLeft,
                              DOUBLE YScale);
static void PS_YSetMinNumbers (FLOAT Start, 
                              SHORT NumberDistanceFromTic, 
                              FLOAT Offset, 
                              SHORT TextRight, 
                              SHORT TextLeft,
                              DOUBLE YScale);

/*********************************************************************************/
/* Private Module Variables */
/*********************************************************************************/
static FILE *m_PSFileStream;

static FLOAT m_Scale;
static PS_COORDINATE_SYSTEM m_MarkerMovementUnits;
static PS_COORDINATE_SYSTEM m_TextMovementUnits;
static FLOAT m_MarkerStrokeGray;
static FLOAT m_MarkerFillGray;
static FLOAT m_MarkerStrokeRed;
static FLOAT m_MarkerFillRed;
static FLOAT m_MarkerStrokeGreen;
static FLOAT m_MarkerFillGreen;
static FLOAT m_MarkerStrokeBlue;
static FLOAT m_MarkerFillBlue;
static BYTE m_MarkerStrokeGrayColor[PS_MAX_STRING_LENGTH];
static BYTE m_MarkerFillGrayColor[PS_MAX_STRING_LENGTH];
static BOOLEAN m_MarkerLoaded[PS_MARKER_MAX_NUMBER];
static BOOLEAN m_UsingColor;
static BOOLEAN m_UsingOpaqueBG;
static BOOLEAN m_XAxisLog;
static BOOLEAN m_YAxisLog;
static BOOLEAN m_AxisEnable;
static BOOLEAN m_AxisNumberEnable;
static USHORT m_LogNumberSelectMask;
static USHORT m_LogTicSelectMask;
static UBYTE m_LogNumberDistanceFromTic;
static PS_AXIS_TIC_AND_NUMBER_TYPE m_LogTicNumberPosition;
static UBYTE m_LogTicLength;
static UBYTE m_LogTicLength10;
static UBYTE m_Precision;
static UBYTE m_LinearTicOffsetNumbering;
static UBYTE m_LinearNumberDistanceFromTic;
static PS_AXIS_TIC_AND_NUMBER_TYPE m_LinearTicNumberPosition;
static UBYTE m_LinearTicLength;
static FLOAT m_RealXMax;
static FLOAT m_RealYMax;
static FLOAT m_XMax;
static FLOAT m_YMax;
static FLOAT m_XMin;
static FLOAT m_YMin;
static FLOAT m_YPositionOfXAxis;
static FLOAT m_XPositionOfYAxis;
static BYTE m_StringAtHighestY[PS_MAX_STRING_LENGTH];
static DOUBLE m_LowestDecadeX;
static DOUBLE m_LowestDecadeY;
static DOUBLE m_UpperDecadeX;
static DOUBLE m_UpperDecadeY;

static FLOAT m_OriginX;
static FLOAT m_OriginY;
static FLOAT m_BGGray;
static FLOAT m_BGRed;
static FLOAT m_BGGreen;
static FLOAT m_BGBlue;
static FLOAT m_FontSize;

static const BYTE *m_FontNameList[PS_NUMBER_PSFONTS] = 
{
   "Helvetica",
   "Helvetica-Bold",
   "Helvetica-Oblique",
   "Helvetica-BoldOblique",
   "Times-Roman",
   "Times-Bold",
   "Times-Italic",
   "Times-BoldItalic",
   "Courier",
   "Courier-Bold",
   "Courier-Oblique",
   "Courier-BoldOblique",
   "Symbol",
   "AvantGarde-Book",
   "AvantGarde-BookOblique",
   "AvantGarde-Demi",
   "AvantGarde-DemiOblique",
   "Bookman-Demi",
   "Bookman-DemiItalic",
   "Bookman-Light",
   "Bookman-LightItalic",
   "Helvetica-Narrow",
   "Helvetica-Narrow-Bold",
   "Helvetica-Narrow-Oblique",
   "Helvetica-Narrow-BoldOblique",
   "NewCenturySchlbk-Roman",
   "NewCenturySchlbk-Italic",
   "NewCenturySchlbk-Bold",
   "NewCenturySchlbk-BoldItalic",
   "Palatino-Roman",
   "Palatino-Italic",
   "Palatino-Bold",
   "Palatino-BoldItalic",
   "ZapfChancery-MediumItalic",
   "ZapfDingbats"
};

/*********************************************************************************/
/* Public and Private Module Functions */
/*********************************************************************************/

void PS_SetDefaults (FLOAT Scale)
{
   UBYTE i;
   
   m_Scale = Scale * PS_POINTS_PER_INCH;
   m_MarkerMovementUnits = PS_DOMAIN;
   m_TextMovementUnits = PS_INCHES;
   for (i = 0; i < PS_MARKER_MAX_NUMBER; i++)
   {
      m_MarkerLoaded[i] = FALSE;
   }
   
   m_OriginX = 0;
   m_OriginY = 0;

   m_BGGray = 1.0;
   m_BGRed = 0.0;
   m_BGGreen = 0.0;
   m_BGBlue = 0.5;

   m_UsingColor = TRUE;
	m_UsingOpaqueBG = FALSE;

   PS_SetMarkerGrayRGBColor (0.0, 0.0, 0.0, 0.0, 1.0, 1.0, 1.0, 1.0);

   m_XAxisLog = FALSE;
   m_YAxisLog = FALSE;
   m_AxisEnable = TRUE;
   m_AxisNumberEnable = TRUE;
   
   m_LogNumberSelectMask = 0x0003;
   m_LogTicSelectMask = 0x03FF;
   m_LogNumberDistanceFromTic = 2;
   m_LogTicNumberPosition = PS_AXIS_TIC_AND_NUMBER_TIC_THROUGH_NUM_LEFT_BELOW;
   m_LogTicLength = 5;
   m_LogTicLength10 = 8;
   m_Precision = 4;
   m_LinearTicOffsetNumbering = 0;
   m_LinearNumberDistanceFromTic = 2;
   m_LinearTicNumberPosition = PS_AXIS_TIC_AND_NUMBER_TIC_THROUGH_NUM_LEFT_BELOW;
   m_LinearTicLength = 5;

   m_YPositionOfXAxis = 10.0;
   m_XPositionOfYAxis = 10.0;
}

/*********************************************************************************/

void PS_Initialize (BOOLEAN UseLandscape, BYTE *TopCommentString, USHORT Width, USHORT Height)
{
   time_t CurrentTime;
   BYTE CurrentTimeString[PS_MAX_STRING_LENGTH];
   BYTE *TailorCommentString = NULL;
   USHORT i;
   USHORT j;
   
   CurrentTime = time(NULL); 
   sprintf(CurrentTimeString, "%s", ctime(&CurrentTime));
   CurrentTimeString[strlen(CurrentTimeString) - 1] = 0;

   m_PSFileStream = fopen(PARMS_GetPostScriptOutputFileName(), "w");
   if (m_PSFileStream == NULL)
   {
      printf("Cannot create %s.\n", PARMS_GetPostScriptOutputFileName());
      exit(1);
   }

   fprintf(m_PSFileStream, "%%!PS-Adobe-3.0 EPSF-3.0\n");   
   fprintf(m_PSFileStream, "%%%%BoundingBox: 0 0 %d %d\n", Width, Height);
   fprintf(m_PSFileStream, "%%%%Title: ICYK (version %d.%d)\n", VERSION_MAJOR, VERSION_MINOR);
   fprintf(m_PSFileStream, "%%%%Creator: ICYK\n");
   fprintf(m_PSFileStream, "%%%%CreationDate: %s\n", CurrentTimeString);
   fprintf(m_PSFileStream, "%%%%For: (empty)\n");

   if (UseLandscape == TRUE)
   {
      fprintf(m_PSFileStream, "%%%%Orientation: Landscape\n");
      fprintf(m_PSFileStream, "%%%%Pages: 1\n%%%%Page: 1 1\n%%%%EndComments\n\n");
      fprintf(m_PSFileStream, "0 %d moveto\ncurrentpoint translate\n", Height);
      fprintf(m_PSFileStream, "270 rotate\n\n");
   }
   else
   {
      fprintf(m_PSFileStream, "%%%%Pages: 1\n%%%%Page: 1 1\n%%%%EndComments\n\n");
   }

   if(TopCommentString != NULL)
   {
      fprintf(m_PSFileStream, "%s", TopCommentString);

      TailorCommentString = (BYTE *) malloc(strlen(TopCommentString) * 2);

      if (TailorCommentString == NULL)
      {
         printf("Error: malloc failed in PS_Initialize.\n");
         exit(1);
      }

      i = 0;
      j = 0;
      while (TopCommentString[i] != 0)
      {
         if ((TopCommentString[i] == '\\') ||
             (TopCommentString[i] == '(') ||
             (TopCommentString[i] == ')'))
         {
            TailorCommentString[j++] = '\\';
         }
         TailorCommentString[j++] = TopCommentString[i++];
      }
      TailorCommentString[j] = 0;
	}

   fprintf(m_PSFileStream, "\n\n%%%%BeginResource: (ICYK; version %d.%d)\n\n",  VERSION_MAJOR, VERSION_MINOR);
   if (m_UsingColor == TRUE)
   {
      fprintf(m_PSFileStream, "/__UseColor true def\n\n");
   }
   else
   {
      fprintf(m_PSFileStream, "/__UseColor false def\n\n");
   }
   
   if (m_UsingOpaqueBG == TRUE)
   {
      fprintf(m_PSFileStream, "/__OpaqueBG true def\n\n");
   }
   else
   {
      fprintf(m_PSFileStream, "/__OpaqueBG false def\n\n");
   }

   fprintf(m_PSFileStream, "/xs {72 mul} def\n/ys {72 mul} def\n");
   fprintf(m_PSFileStream, "/cxs {cvr xs} def\n/cys {cvr ys} def\n");
   fprintf(m_PSFileStream, "/cp {closepath} def\n");
   fprintf(m_PSFileStream, "/gs {gsave} def\n");
   fprintf(m_PSFileStream, "/gr {grestore} def\n");
   fprintf(m_PSFileStream, "/ff {findfont} def\n");
   fprintf(m_PSFileStream, "/stw {stringwidth} def\n");
   fprintf(m_PSFileStream, "/ssf {scalefont setfont} def\n");
   fprintf(m_PSFileStream, "/cpt {currentpoint translate} def\n");
   fprintf(m_PSFileStream, "/rlto {rlineto} def\n");
   fprintf(m_PSFileStream, "/rmto {rmoveto} def\n");
   fprintf(m_PSFileStream, "/lto {lineto} def\n");
   fprintf(m_PSFileStream, "/mto {moveto} def\n");
   fprintf(m_PSFileStream, "/setg {setgray} def\n");
   fprintf(m_PSFileStream, "/slw {setlinewidth} def\n");
   fprintf(m_PSFileStream, "/savlw {/clw currentlinewidth def} def\n");
   fprintf(m_PSFileStream, "/sl {cvr ys exch cvr xs exch lto} def\n");
   fprintf(m_PSFileStream, "/sm {cvr ys exch cvr xs exch mto} def\n");
   fprintf(m_PSFileStream, "/nd {[] 0 setdash} def\n");
   fprintf(m_PSFileStream, "/charheight {newpath 0 0 mto (I) false\n");
   fprintf(m_PSFileStream, "	charpath flattenpath pathbbox\n");
   fprintf(m_PSFileStream, "	exch pop exch sub exch pop newpath} def\n");
   fprintf(m_PSFileStream, "__UseColor { /setgrayrgbcolor { setrgbcolor pop } def }\n");
   fprintf(m_PSFileStream, "    { /setgrayrgbcolor { pop pop pop setgray } def } ifelse\n");
   fprintf(m_PSFileStream, "%%%%EndResource\n\n");

   fprintf(m_PSFileStream, "%%%%BeginResource: procset (Tailor group constructs)\n");
   fprintf(m_PSFileStream, "/TailorGroupBegin {pop} def /TailorGroupEnd {} def\n");
   fprintf(m_PSFileStream, "/pdfmark where\n");
   fprintf(m_PSFileStream, "{pop} {userdict /pdfmark /cleartomark load put} ifelse\n\n");
   fprintf(m_PSFileStream, "%%%%EndResource\n\n");

   fprintf(m_PSFileStream, "%%%%EndProlog\n\n");

   fprintf(m_PSFileStream, "(%% Generated by: ICYK\n");
   if(TailorCommentString != NULL)
   {
      fprintf(m_PSFileStream, "%s", TailorCommentString);
	}
   fprintf(m_PSFileStream,") TailorGroupBegin\n\n");

   /* 
      Insert topcomments as an annotation pdfmark.  Comments will then become
      an annotation in the Acrobat (PDF) file once the file is distilled by
      an Acrobat Distiller.  It has no effect as long as the file remains in PS.
   */
   if(TailorCommentString != NULL) 
   {
      fprintf(m_PSFileStream, "%% Annotation for Acrobat (PDF) distiller\n");
      fprintf(m_PSFileStream, "[ /Rect [0 0 300 150]\n");
      fprintf(m_PSFileStream, "/Contents (");
      fprintf(m_PSFileStream, "%s)\n", TailorCommentString);
      fprintf(m_PSFileStream, "/Title (ICYK plot on %s)\n", CurrentTimeString);
      fprintf(m_PSFileStream, "/Open false\n");       /* initialize for iconized appearance */
      fprintf(m_PSFileStream, "/ANN pdfmark\n\n");
   }

   fprintf(m_PSFileStream, "%% Executable PS code begins here.\n\n");
   fprintf(m_PSFileStream, "%% Background drawn here if enabled above. Args:(gray r g b)\n");
   fprintf(m_PSFileStream, "__OpaqueBG { %g %g %g %g setgrayrgbcolor clippath fill } if \n\n", m_BGGray, m_BGRed, m_BGGreen, m_BGBlue);

   fprintf(m_PSFileStream, "0 0 0 0 setgrayrgbcolor 1 slw\n");
   fprintf(m_PSFileStream, "1 setlinecap 0 setlinejoin\n");
   fprintf(m_PSFileStream, "%% End of stuff generated by PS_Initialize()\n\n");
   
   PS_Font(12.0, PS_FONT_HELVETICA, PS_FONT_STYLE_NORMAL);

   if(TailorCommentString != NULL)
   {
      free(TailorCommentString);
   }
}

/*********************************************************************************/

void PS_ShowPage (void) 
{
   fprintf(m_PSFileStream, "\n\n%% All drawing done.  Commit to imaging device.\n");
   fprintf(m_PSFileStream, "showpage\n");
   fprintf(m_PSFileStream, "TailorGroupEnd\n\n");
   fprintf(m_PSFileStream, "%%%%Trailer\n");

   fclose(m_PSFileStream);
}

/*********************************************************************************/

void PS_SetDashPattern (PS_LINE_DASH_TYPE Type, FLOAT DashPeriod) 
{
   fprintf(m_PSFileStream, "\n\n%% Setting dash pattern\n");
   if ((DashPeriod < 0.01) || (DashPeriod > 20.0)) 
   {
      printf("Error: Dash period %g is out of range (0.01 .. 20.0).\n", DashPeriod);
      exit(1);
   }
   
   switch(Type) 
   {
      /* 
         Keep sum of all dash components to be 72 points. 
      */
      case PS_LINE_NO_DASH: 
         fprintf(m_PSFileStream, "nd\n");
         break;
         
      case PS_LINE_DASH_1: 
         fprintf(m_PSFileStream, "[%.5g %.5g] 0 setdash\n", 
                                 (62.0 * DashPeriod), (10.0 * DashPeriod));
         break;
          
      case PS_LINE_DASH_2: 
         fprintf(m_PSFileStream, "[%.5g %.5g %.5g %.5g %.5g %.5g] 0 setdash\n", 
                                 (30.0 * DashPeriod), (10.0 * DashPeriod), 
                                 (6.0 * DashPeriod), (10.0 * DashPeriod), 
                                 (6.0 * DashPeriod), (10.0 * DashPeriod));
         break;
          
      case PS_LINE_DASH_3: 
         fprintf(m_PSFileStream, "[%.5g %.5g %.5g %.5g] 0 setdash\n",
                                 (36.0 * DashPeriod), (12.0 * DashPeriod), 
                                 (12.0 * DashPeriod), (12.0 * DashPeriod));
         break;
          
      case PS_LINE_DASH_4: 
         fprintf(m_PSFileStream, "[%.5g %.5g] 0 setdash\n", 
                                 (54.0 * DashPeriod), (18.0 * DashPeriod));
         break;
          
      case PS_LINE_DASH_5: 
         fprintf(m_PSFileStream, "[%.5g %.5g] 0 setdash\n", 
                                 (18.0 * DashPeriod), (54.0 * DashPeriod));
         break;

      case PS_LINE_DASH_6: 
         fprintf(m_PSFileStream, "[%.5g %.5g] 0 setdash\n", 
                                 (36.0 * DashPeriod), (36.0 * DashPeriod));
         break;
   }
   
	fprintf(m_PSFileStream, "%% END Setting dash pattern\n\n");
}

/*********************************************************************************/

void PS_CenterLabel (BYTE *TextString, FLOAT X, FLOAT Y , FLOAT RotateBy, PS_COORDINATE_SYSTEM CoordinateSystem)
{
   fprintf(m_PSFileStream, "\n\n%% Centered text label\n");

   fprintf(m_PSFileStream, "gsave\n");
   if (CoordinateSystem == PS_DOMAIN)
   {
      fprintf(m_PSFileStream, " %g %g sm\t ", X, Y);
   }
   else
   {		
      fprintf(m_PSFileStream, " %g %g mto\t ", (X * m_Scale), (Y * m_Scale));
   }
   fprintf(m_PSFileStream, " currentpoint translate ");
   fprintf(m_PSFileStream, "%g rotate ", RotateBy);
   fprintf(m_PSFileStream, "(%s) stringwidth ", TextString);
   fprintf(m_PSFileStream, "2 div -1 mul exch 2 div \n -1 mul exch rmoveto ");
   fprintf(m_PSFileStream, "(%s) show ", TextString);
   fprintf(m_PSFileStream, "grestore\n");
   fprintf(m_PSFileStream, "%% END Centered text label\n\n");
}

/*********************************************************************************/

void PS_XLabel (BYTE *TextString)
{
   PS_CenterLabel (TextString, m_RealXMax /2., (m_YPositionOfXAxis - (0.4 * m_FontSize)) / 72.0, 0.0, PS_INCHES);
}

/*********************************************************************************/

void PS_YLabel (BYTE *TextString)
{
   fprintf(m_PSFileStream, "\n\n%% Centered Y label\n");

   fprintf(m_PSFileStream, "gsave\n");
   fprintf(m_PSFileStream, " %g %g ", (0.5 * m_RealYMax * m_Scale), m_XPositionOfYAxis);	
   fprintf(m_PSFileStream, " (%s) stringwidth ", m_StringAtHighestY);
   fprintf(m_PSFileStream, " pop sub ");
   fprintf(m_PSFileStream, " %g sub exch", 0.4 * m_FontSize); 
   fprintf(m_PSFileStream, " moveto currentpoint translate ");
   fprintf(m_PSFileStream, " %g rotate ", 90.0);
   fprintf(m_PSFileStream, " (%s) stringwidth ", TextString);
   fprintf(m_PSFileStream, " exch 2 div -1 mul exch -1 mul rmoveto ");
   fprintf(m_PSFileStream, "(%s) show ", TextString);
   fprintf(m_PSFileStream, "grestore\n");
   fprintf(m_PSFileStream, "%% END Centered Y label\n\n");
}

/*********************************************************************************/

void PS_TextAlign (FLOAT X, FLOAT Y, FLOAT RotateBy, BYTE *TextString, SHORT XPos, SHORT YPos, PS_COORDINATE_SYSTEM CoordinateSystem)
{
   fprintf(m_PSFileStream, "\n\n%% Aligned text label\n");
   fprintf(m_PSFileStream, "gsave\n");
   if (CoordinateSystem == PS_DOMAIN)
   {
      fprintf(m_PSFileStream, " %g %g sm\t ", X, Y);
   }
   else
   {		
      fprintf(m_PSFileStream, " %g %g mto\t ", (X * m_Scale), (Y * m_Scale));
   }
   fprintf(m_PSFileStream, " currentpoint translate ");
   fprintf(m_PSFileStream, " %g rotate", RotateBy);
   fprintf(m_PSFileStream, " (%s) stringwidth pop 2 div %g mul ", TextString, (-1.0 * ((FLOAT) XPos)));
   fprintf(m_PSFileStream, " %g ", (-1.0 * (m_FontSize / 2.7) * ((FLOAT) YPos))); 
   fprintf(m_PSFileStream, " rmoveto ");	
   fprintf(m_PSFileStream, " (%s) show ", TextString);
   fprintf(m_PSFileStream, "grestore\n");
   fprintf(m_PSFileStream, "%% END Aligned text label\n\n");
}

/*********************************************************************************/

void PS_SelectCoordinate (PS_COORDINATE_SYSTEM Marks, PS_COORDINATE_SYSTEM Text)
{
   m_MarkerMovementUnits = Marks;
   m_TextMovementUnits = Text;
}

/*********************************************************************************/

void PS_Font (FLOAT Size, PS_FONT_TYPE Type, PS_FONT_STYLE Style)
{
   m_FontSize = Size;

   switch(Type)
   {
      case PS_FONT_TIMES:
         fprintf(m_PSFileStream, "/Times-");
         switch(Style)
         {
            case PS_FONT_STYLE_NORMAL:
               fprintf(m_PSFileStream, "Roman");
               break;

            case PS_FONT_STYLE_BOLD:
               fprintf(m_PSFileStream, "Bold");
               break;

            case PS_FONT_STYLE_ITALIC:
               fprintf(m_PSFileStream, "Italic");
               break;

            case PS_FONT_STYLE_BOLD_ITALIC:
               fprintf(m_PSFileStream, "BoldItalic");
               break;
         }
         break;
         
      case PS_FONT_COURIER:
         fprintf(m_PSFileStream, "/Courier");
         switch(Style)
         {
            case PS_FONT_STYLE_NORMAL:
               break;

            case PS_FONT_STYLE_BOLD:
               fprintf(m_PSFileStream, "-Bold");
               break;

            case PS_FONT_STYLE_ITALIC:
               fprintf(m_PSFileStream, "-Oblique");
               break;

            case PS_FONT_STYLE_BOLD_ITALIC:
               fprintf(m_PSFileStream, "-BoldOblique");
               break;
         }
         break;

      case PS_FONT_HELVETICA:
         fprintf(m_PSFileStream, "/Helvetica");
         switch(Style)
         {
            case PS_FONT_STYLE_NORMAL:
               break;

            case PS_FONT_STYLE_BOLD:
               fprintf(m_PSFileStream, "-Bold");
               break;

            case PS_FONT_STYLE_ITALIC:
               fprintf(m_PSFileStream, "-Oblique");
               break;

            case PS_FONT_STYLE_BOLD_ITALIC:
               fprintf(m_PSFileStream, "-BoldOblique");
               break;
         }
         break;
   }
   
   fprintf(m_PSFileStream, " ff %g ssf\n", m_FontSize);
}

/*********************************************************************************/

void PS_FontByName (FLOAT Size, BYTE *Name)
{
   UBYTE i;
   BOOLEAN FontOkay = FALSE;
	
   for(i = 0; i < PS_NUMBER_PSFONTS; i++)
   {
      if(strcmp(Name, m_FontNameList[i]) == 0)
      {
         FontOkay = TRUE;
      }
   }
   
   if(FontOkay == FALSE) 
   {
      printf("Error: bad font name.\n");
      exit(1);
   }

   m_FontSize = Size;

   fprintf(m_PSFileStream, "/%s ff %g ssf\n", Name, Size);
}

/*********************************************************************************/

void PS_SetGray (FLOAT Degree) 	
{
   fprintf(m_PSFileStream, "\n%g setgray\n", Degree);
}

/*********************************************************************************/

void PS_SetRGBColor (FLOAT Red, FLOAT Green, FLOAT Blue)
{
   fprintf(m_PSFileStream, "\n%g %g %g setrgbcolor\n", Red, Green, Blue);
}

/*********************************************************************************/

void PS_SetGrayRGBColor(FLOAT Degree, FLOAT Red, FLOAT Green, FLOAT Blue)
{
   fprintf(m_PSFileStream, "\n%g %g %g %g setgrayrgbcolor\n", Degree, Red, Green, Blue);
}

/*********************************************************************************/

void PS_SetMarkerGray (FLOAT NewPerimeter, FLOAT NewInside)
{
   UBYTE i;

   m_MarkerStrokeGray = NewPerimeter;
   m_MarkerFillGray = NewInside;
   sprintf(m_MarkerStrokeGrayColor, " %g setgray ", m_MarkerStrokeGray);
   sprintf(m_MarkerFillGrayColor, " %g setgray ", m_MarkerFillGray);
   for (i = 0; i < PS_MARKER_MAX_NUMBER; i++)
   {
      m_MarkerLoaded[i] = FALSE;
   }
}

/*********************************************************************************/

void PS_SetMarkerColor (FLOAT NewPerimeterRed, 
                        FLOAT NewPerimeterGreen, 
                        FLOAT NewPerimeterBlue, 
                        FLOAT NewInsideRed, 
                        FLOAT NewInsideGreen, 
                        FLOAT NewInsideBlue)
{
   UBYTE i;

   m_MarkerStrokeRed = NewPerimeterRed;
   m_MarkerFillRed = NewInsideRed;
   m_MarkerStrokeGreen = NewPerimeterGreen;
   m_MarkerFillGreen = NewInsideGreen;
   m_MarkerStrokeBlue = NewPerimeterBlue;
   m_MarkerFillBlue = NewInsideBlue;
   sprintf(m_MarkerStrokeGrayColor, " %g %g %g setrgbcolor ", m_MarkerStrokeRed, m_MarkerStrokeGreen, m_MarkerStrokeBlue);
   sprintf(m_MarkerFillGrayColor, " %g %g %g setrgbcolor ", m_MarkerFillRed, m_MarkerFillGreen, m_MarkerFillBlue);
   for (i = 0; i < PS_MARKER_MAX_NUMBER; i++)
   {
      m_MarkerLoaded[i] = FALSE;
   }
}

/*********************************************************************************/

void PS_SetMarkerGrayRGBColor (FLOAT NewPerimeterGray, 
                        FLOAT NewPerimeterRed, 
                        FLOAT NewPerimeterGreen, 
                        FLOAT NewPerimeterBlue, 
                        FLOAT NewInsideGray, 
                        FLOAT NewInsideRed, 
                        FLOAT NewInsideGreen, 
                        FLOAT NewInsideBlue)
{
	UBYTE i;

   m_MarkerStrokeGray = NewPerimeterGray;
   m_MarkerFillGray = NewInsideGray;
   m_MarkerStrokeRed = NewPerimeterRed;
   m_MarkerFillRed = NewInsideRed;
   m_MarkerStrokeGreen = NewPerimeterGreen;
   m_MarkerFillGreen = NewInsideGreen;
   m_MarkerStrokeBlue = NewPerimeterBlue;
   m_MarkerFillBlue = NewInsideBlue;
   sprintf(m_MarkerStrokeGrayColor, " %g %g %g %g setgrayrgbcolor ", m_MarkerStrokeGray, m_MarkerStrokeRed, m_MarkerStrokeGreen, m_MarkerStrokeBlue);
   sprintf(m_MarkerFillGrayColor, " %g %g %g %g setgrayrgbcolor ", m_MarkerFillGray, m_MarkerFillRed, m_MarkerFillGreen, m_MarkerFillBlue);
   for (i = 0; i < PS_MARKER_MAX_NUMBER; i++)
   {
      m_MarkerLoaded[i] = FALSE;
   }
}

/*********************************************************************************/

void PS_UseColor (BOOLEAN Flag)
{
   m_UsingColor = Flag;
}

/*********************************************************************************/

void PS_UseOpaqueBG (BOOLEAN Flag, FLOAT Gray, FLOAT Red, FLOAT Green, FLOAT Blue)
{
   m_BGGray = Gray;
   m_BGRed = Red;
   m_BGGreen = Green;
   m_BGBlue = Blue;
	m_UsingOpaqueBG = Flag;
}

/*********************************************************************************/

void PS_Marker(FLOAT X, FLOAT Y, PS_MARKER_TYPE Type, FLOAT Size)
{
   fprintf(m_PSFileStream, "\ngs newpath nd\n");
   fprintf(m_PSFileStream, "/size %g def\n", Size);
   if (m_MarkerMovementUnits == PS_DOMAIN)
   {
      fprintf(m_PSFileStream, "/x %g xs def\t/y %g ys def\n", X, Y);
   }
   else
   {		
      fprintf(m_PSFileStream, "/x %g def\t/y %g def\n", (X * m_Scale), (Y * m_Scale));
   }

   if (m_MarkerLoaded[Type] == FALSE)
   {
      PS_LoadMark (Type);
      m_MarkerLoaded[Type] = TRUE;
   }
   
   switch (Type)
   {
      case PS_MARKER_OPEN_CIRCLE:
         fprintf(m_PSFileStream, "WCircle ");
         break;
      
      case PS_MARKER_FILLED_CIRCLE:
         fprintf(m_PSFileStream, "BCircle ");
         break;
      
      case PS_MARKER_OPEN_TRIANGLE:
         fprintf(m_PSFileStream, "Wtriangle ");
         break;
   
      case PS_MARKER_FILLED_TRIANGLE:
         fprintf(m_PSFileStream, "Btriangle ");
         break;
   
      case PS_MARKER_OPEN_DIAMOND:
         fprintf(m_PSFileStream, "Wdiamond ");
         break;

      case PS_MARKER_FILLED_DIAMOND:
         fprintf(m_PSFileStream, "Bdiamond ");
         break;

      case PS_MARKER_OPEN_SQUARE:
         fprintf(m_PSFileStream, "Wsquare ");
         break;

      case PS_MARKER_FILLED_SQUARE:
         fprintf(m_PSFileStream, "Bsquare ");
         break;

      case PS_MARKER_OPEN_STAR:
         fprintf(m_PSFileStream, "Wstar ");
         break;

      case PS_MARKER_FILLED_STAR:
         fprintf(m_PSFileStream, "Bstar ");
         break;

      case PS_MARKER_OPEN_PENTAGON:
         fprintf(m_PSFileStream, "Wpent ");
         break;

      case PS_MARKER_FILLED_PENTAGON:
         fprintf(m_PSFileStream, "Bpent ");
         break;

      case PS_MARKER_OPEN_INVERTED_TRIANGLE:
         fprintf(m_PSFileStream, "Wintri ");
         break;

      case PS_MARKER_FILLED_INVERTED_TRIANGLE:
         fprintf(m_PSFileStream, "Bintri ");
         break;

      case PS_MARKER_BOWTIE:
         fprintf(m_PSFileStream, "Bowtie ");
         break;

      case PS_MARKER_JIGGER:
         fprintf(m_PSFileStream, "Jigger ");
         break;

      case PS_MARKER_ASTERISK:
         fprintf(m_PSFileStream, "Asterisk ");
         break;

      case PS_MARKER_CROSSHAIRS:
         fprintf(m_PSFileStream, "Xhairs ");
         break;

      case PS_MARKER_SPINNING_WHEEL:
         fprintf(m_PSFileStream, "Spinner ");
         break;

      case PS_MARKER_CROSSHATCH:
         fprintf(m_PSFileStream, "Xhatch ");
         break;

      default:
         break;
   }
   fprintf(m_PSFileStream, " gr\n");
}

/*********************************************************************************/

void PS_LoadMark(PS_MARKER_TYPE Type)
{
   switch(Type)
   {
      case PS_MARKER_OPEN_CIRCLE:
         fprintf(m_PSFileStream, "/WCircle {\n");
         fprintf(m_PSFileStream, " size 8 div slw\n");
         fprintf(m_PSFileStream, " x y size .5 mul 0 360 arc closepath\n");
         fprintf(m_PSFileStream, " gs %s fill gr\n", m_MarkerFillGrayColor);
         fprintf(m_PSFileStream, " %s stroke} def\n\n", m_MarkerStrokeGrayColor);
         break;

      case PS_MARKER_FILLED_CIRCLE:
         fprintf(m_PSFileStream, "/BCircle {\n");
         fprintf(m_PSFileStream, " size 8 div slw\n");
         fprintf(m_PSFileStream, " x y size .5 mul 0 360 arc closepath\n");
         fprintf(m_PSFileStream, " gs %s fill gr\n", m_MarkerStrokeGrayColor);
         fprintf(m_PSFileStream, " %s stroke} def\n\n", m_MarkerStrokeGrayColor);
         break;

      case PS_MARKER_OPEN_TRIANGLE:
         fprintf(m_PSFileStream, "/Wtriangle {\n");
         fprintf(m_PSFileStream, " size 8 div slw\n");
         fprintf(m_PSFileStream, " x .667 size mul y add mto\n");
         fprintf(m_PSFileStream, " size neg 1.7321 div size neg rlto\n");
         fprintf(m_PSFileStream, " size .8661 div 0 rlto\n");
         fprintf(m_PSFileStream, " cp gs %s fill gr\n", m_MarkerFillGrayColor);
         fprintf(m_PSFileStream, " %s stroke} def\n\n", m_MarkerStrokeGrayColor);
         break;

      case PS_MARKER_FILLED_TRIANGLE:
         fprintf(m_PSFileStream, "/Btriangle {\n");
         fprintf(m_PSFileStream, " size 8 div slw\n");
         fprintf(m_PSFileStream, " x .667 size mul y add mto\n");
         fprintf(m_PSFileStream, " size neg 1.7321 div size neg rlto\n");
         fprintf(m_PSFileStream, " size .8661 div 0 rlto\n");
         fprintf(m_PSFileStream, " cp gs %s fill gr\n", m_MarkerStrokeGrayColor);
         fprintf(m_PSFileStream, " %s stroke} def\n\n", m_MarkerStrokeGrayColor);
         break;

      case PS_MARKER_OPEN_DIAMOND:
         fprintf(m_PSFileStream, "/Wdiamond {\n");
         fprintf(m_PSFileStream, " size 8 div slw\n");
         fprintf(m_PSFileStream, " x size 1.38 div y add mto\n");
         fprintf(m_PSFileStream, " size .546 mul size 1.38 div neg rlto\n");
         fprintf(m_PSFileStream, " size .546 mul neg size 1.38 div neg rlto\n");
         fprintf(m_PSFileStream, " size .546 mul neg size 1.38 div  rlto\n");
         fprintf(m_PSFileStream, " cp gs %s fill gr\n", m_MarkerFillGrayColor);
         fprintf(m_PSFileStream, " %s stroke} def\n\n", m_MarkerStrokeGrayColor);
         break;

      case PS_MARKER_FILLED_DIAMOND:
         fprintf(m_PSFileStream, "/Bdiamond {\n");
         fprintf(m_PSFileStream, " size 8 div slw\n");
         fprintf(m_PSFileStream, " x size 1.38 div y add mto\n");
         fprintf(m_PSFileStream, " size .546 mul size 1.38 div neg rlto\n");
         fprintf(m_PSFileStream, " size .546 mul neg size 1.38 div neg rlto\n");
         fprintf(m_PSFileStream, " size .546 mul neg size 1.38 div  rlto\n");
         fprintf(m_PSFileStream, " cp gs %s fill gr\n", m_MarkerStrokeGrayColor);
         fprintf(m_PSFileStream, " %s stroke} def\n\n", m_MarkerStrokeGrayColor);
         break;

      case PS_MARKER_OPEN_SQUARE:
         fprintf(m_PSFileStream, "/Wsquare {\n");
         fprintf(m_PSFileStream, " size 8 div slw\n");
         fprintf(m_PSFileStream, " size 2 div dup x add exch y\n");
         fprintf(m_PSFileStream, " add mto 0 size neg rlto size\n"); 
         fprintf(m_PSFileStream, " neg 0 rlto 0 size rlto cp gs\n");
         fprintf(m_PSFileStream, " %s fill gr\n", m_MarkerFillGrayColor); 
         fprintf(m_PSFileStream, " %s stroke} def\n\n", m_MarkerStrokeGrayColor);
         break;

      case PS_MARKER_FILLED_SQUARE:
         fprintf(m_PSFileStream, "/Bsquare {\n");
         fprintf(m_PSFileStream, " size 8 div slw\n");
         fprintf(m_PSFileStream, " size 2 div dup x add exch y\n");
         fprintf(m_PSFileStream, " add mto 0 size neg rlto size\n"); 
         fprintf(m_PSFileStream, " neg 0 rlto 0 size rlto cp gs\n");
         fprintf(m_PSFileStream, " %s fill gr\n", m_MarkerStrokeGrayColor); 
         fprintf(m_PSFileStream, " %s stroke} def\n\n", m_MarkerStrokeGrayColor);
         break;

      case PS_MARKER_OPEN_STAR:
         fprintf(m_PSFileStream, "/Wstar {\n");
         fprintf(m_PSFileStream, " size 4 div slw\n");
         fprintf(m_PSFileStream, " .5257 size mul neg x add\n");
         fprintf(m_PSFileStream, " .1708 size mul y add mto\n");
         fprintf(m_PSFileStream, " 1.0515 size mul 0 rlto\n");
         fprintf(m_PSFileStream, " size .8507 mul neg size .6181 mul neg rlto\n");
         fprintf(m_PSFileStream, " .3249 size mul size rlto\n");
         fprintf(m_PSFileStream, " .3249 size mul size neg rlto\n");
         fprintf(m_PSFileStream, " cp\n");
         fprintf(m_PSFileStream, " gs %s stroke gr %s fill} def\n\n", m_MarkerStrokeGrayColor, m_MarkerFillGrayColor);
         break;

      case PS_MARKER_FILLED_STAR:
         fprintf(m_PSFileStream, "/Bstar {\n");
         fprintf(m_PSFileStream, " size 4 div slw\n");
         fprintf(m_PSFileStream, " .5257 size mul neg x add\n");
         fprintf(m_PSFileStream, " .1708 size mul y add mto\n");
         fprintf(m_PSFileStream, " 1.0515 size mul 0 rlto\n");
         fprintf(m_PSFileStream, " size .8507 mul neg size .6181 mul neg rlto\n");
         fprintf(m_PSFileStream, " .3249 size mul size rlto\n");
         fprintf(m_PSFileStream, " .3249 size mul size neg rlto\n");
         fprintf(m_PSFileStream, " cp\n");
         fprintf(m_PSFileStream, " gs %s stroke gr %s fill} def\n\n", m_MarkerStrokeGrayColor, m_MarkerStrokeGrayColor);
         break;
      
      case PS_MARKER_OPEN_PENTAGON:
         fprintf(m_PSFileStream, "/Wpent {\n");
         fprintf(m_PSFileStream, " size 8 div slw\n");
         fprintf(m_PSFileStream, " .5257 size mul neg x add\n");
         fprintf(m_PSFileStream, " .1708 size mul y add mto\n");
         fprintf(m_PSFileStream, " .5257 size mul .382 size mul rlto\n");
         fprintf(m_PSFileStream, " .5257 size mul .382 size neg mul rlto\n");
         fprintf(m_PSFileStream, " size neg dup .2008 mul exch .6181 mul rlto\n");
         fprintf(m_PSFileStream, " size neg .6499 mul 0 rlto\n");
         fprintf(m_PSFileStream, " cp gs %s fill gr\n", m_MarkerFillGrayColor);
         fprintf(m_PSFileStream, " %s stroke} def\n\n", m_MarkerStrokeGrayColor);
         break;

      case PS_MARKER_FILLED_PENTAGON:
         fprintf(m_PSFileStream, "/Bpent {\n");
         fprintf(m_PSFileStream, " size 8 div slw\n");
         fprintf(m_PSFileStream, " .5257 size mul neg x add\n");
         fprintf(m_PSFileStream, " .1708 size mul y add mto\n");
         fprintf(m_PSFileStream, " .5257 size mul .382 size mul rlto\n");
         fprintf(m_PSFileStream, " .5257 size mul .382 size neg mul rlto\n");
         fprintf(m_PSFileStream, " size neg dup .2008 mul exch .6181 mul rlto\n");
         fprintf(m_PSFileStream, " size neg .6499 mul 0 rlto\n");
         fprintf(m_PSFileStream, " cp gs %s fill gr\n", m_MarkerStrokeGrayColor);
         fprintf(m_PSFileStream, " %s stroke} def\n\n", m_MarkerStrokeGrayColor);
         break;
      
      case PS_MARKER_OPEN_INVERTED_TRIANGLE:
         fprintf(m_PSFileStream, "/Wintri {\n");
         fprintf(m_PSFileStream, " size 8 div slw\n");
         fprintf(m_PSFileStream, " x .667 size mul neg y add \n");
         fprintf(m_PSFileStream, " mto size dup 1.7321 div exch rlto\n");
         fprintf(m_PSFileStream, " size neg .8661 div 0 rlto\n");
         fprintf(m_PSFileStream, " cp gs %s fill gr\n", m_MarkerFillGrayColor);
         fprintf(m_PSFileStream, " %s stroke} def\n\n", m_MarkerStrokeGrayColor);
         break;

      case PS_MARKER_FILLED_INVERTED_TRIANGLE:
         fprintf(m_PSFileStream, "/Bintri {\n");
         fprintf(m_PSFileStream, " size 8 div slw\n");
         fprintf(m_PSFileStream, " x .667 size mul neg y add \n");
         fprintf(m_PSFileStream, " mto size dup 1.7321 div exch rlto\n");
         fprintf(m_PSFileStream, " size neg .8661 div 0 rlto\n");
         fprintf(m_PSFileStream, " cp gs %s fill gr\n", m_MarkerStrokeGrayColor);
         fprintf(m_PSFileStream, " %s stroke} def\n\n", m_MarkerStrokeGrayColor);
         break;
      
      case PS_MARKER_BOWTIE:
         fprintf(m_PSFileStream, "/Bowtie {\n");
         fprintf(m_PSFileStream, " size 8 div slw\n");
         fprintf(m_PSFileStream, " size 2 div dup neg x add exch y\n");
         fprintf(m_PSFileStream, " add mto size dup neg rlto 0 size \n");
         fprintf(m_PSFileStream, " rlto size neg dup rlto cp gs\n");
         fprintf(m_PSFileStream, " %s fill gr %s stroke} def\n\n", m_MarkerFillGrayColor, m_MarkerStrokeGrayColor);
         break;

      case PS_MARKER_JIGGER:
         fprintf(m_PSFileStream, "/Jigger {\n");
         fprintf(m_PSFileStream, " size 8 div slw\n");
         fprintf(m_PSFileStream, " size 2 div dup neg x add exch y\n");
         fprintf(m_PSFileStream, " add mto size dup neg rlto size neg\n");
         fprintf(m_PSFileStream, " 0 rlto size dup rlto cp gs\n");
         fprintf(m_PSFileStream, " %s fill gr %s stroke} def\n\n", m_MarkerFillGrayColor, m_MarkerStrokeGrayColor);
         break;

      case PS_MARKER_ASTERISK:
         fprintf(m_PSFileStream, "/Asterisk {\n");
         fprintf(m_PSFileStream, " size 8 div slw\n");
         fprintf(m_PSFileStream, " x size 2 div y add mto\n");
         fprintf(m_PSFileStream, " 0 size neg rlto\n");
         fprintf(m_PSFileStream, " size dup neg .433 mul exch 4 div rmto\n");
         fprintf(m_PSFileStream, " size dup .866 mul exch 2 div rlto\n");
         fprintf(m_PSFileStream, " 0 size neg 2 div rmto\n");
         fprintf(m_PSFileStream, " size dup neg .866 mul exch 2 div rlto\n");
         fprintf(m_PSFileStream, " %s stroke} def\n\n", m_MarkerStrokeGrayColor);
         break;

      case PS_MARKER_CROSSHAIRS:
         fprintf(m_PSFileStream, "/Xhairs {\n");
         fprintf(m_PSFileStream, " size 10 div slw\n");
         fprintf(m_PSFileStream, " x y size 4 div 0 360 arc\n");
         fprintf(m_PSFileStream, " gs %s fill gr %s\n", m_MarkerFillGrayColor, m_MarkerStrokeGrayColor);
         fprintf(m_PSFileStream, " stroke\n");
         fprintf(m_PSFileStream, " x size 4 div y add mto\n");
         fprintf(m_PSFileStream, " 0 size 4 div rlto\n");
         fprintf(m_PSFileStream, " size 2 div dup neg rmto\n");
         fprintf(m_PSFileStream, " size neg 4 div 0 rlto\n");
         fprintf(m_PSFileStream, " size neg 4 div dup rmto\n");
         fprintf(m_PSFileStream, " 0 size neg 4 div rlto\n");
         fprintf(m_PSFileStream, " size 2 div dup neg exch rmto\n");
         fprintf(m_PSFileStream, " size 4 div 0 rlto\n");
         fprintf(m_PSFileStream, " 0 size 2 div rmto\n");
         fprintf(m_PSFileStream, " size 2 div 0 rlto\n");
         fprintf(m_PSFileStream, " size 4 div dup neg rmto\n");
         fprintf(m_PSFileStream, " 0 size neg 2 div rlto\n");
         fprintf(m_PSFileStream, " size neg 4 div dup rmto\n");
         fprintf(m_PSFileStream, " size neg 2 div 0 rlto\n");
         fprintf(m_PSFileStream, " size 4 div dup neg exch rmto\n");
         fprintf(m_PSFileStream, " 0 size 2 div rlto\n");
         fprintf(m_PSFileStream, " %s stroke} def\n\n", m_MarkerStrokeGrayColor);
         break;
      
      case PS_MARKER_SPINNING_WHEEL:
         fprintf(m_PSFileStream, "/Spinner {\n");
         fprintf(m_PSFileStream, " size 10 div slw\n");
         fprintf(m_PSFileStream, " size 4 div dup neg x add exch y\n");
         fprintf(m_PSFileStream, " add mto size 2 div 0 rlto\n");
         fprintf(m_PSFileStream, " 0 size neg 2 div rlto\n");
         fprintf(m_PSFileStream, " size neg 2 div 0 rlto cp\n");
         fprintf(m_PSFileStream, " 1 setg gs fill gr 0 setg\n");
         fprintf(m_PSFileStream, " size 4 div dup neg exch rlto\n");
         fprintf(m_PSFileStream, " size .75 mul dup 3 div neg rlto\n");
         fprintf(m_PSFileStream, " size 4 div dup rlto\n");
         fprintf(m_PSFileStream, " size neg 4 div dup 3 mul rlto\n");
         fprintf(m_PSFileStream, " size 4 div dup neg rlto\n");
         fprintf(m_PSFileStream, " size 4 div dup neg 3 mul exch rlto\n");
         fprintf(m_PSFileStream, " size 4 div neg dup rlto\n");
         fprintf(m_PSFileStream, " size 4 div dup 3 mul rlto %s stroke} def\n\n", m_MarkerStrokeGrayColor);
         break;

      case PS_MARKER_CROSSHATCH:
         fprintf(m_PSFileStream, "/Xhatch {\n");
         fprintf(m_PSFileStream, " size 8 div slw\n");
         fprintf(m_PSFileStream, " size dup neg 4 div x add exch 2\n");
         fprintf(m_PSFileStream, " div y add mto size .75 mul dup\n");
         fprintf(m_PSFileStream, " neg rlto size neg 4 div dup rmto\n");
         fprintf(m_PSFileStream, " size .75 mul dup neg exch rlto\n");
         fprintf(m_PSFileStream, " 0 size neg 2 div rmto\n");
         fprintf(m_PSFileStream, " size .75 mul dup rlto\n");
         fprintf(m_PSFileStream, " size 4 div dup neg rmto\n");
         fprintf(m_PSFileStream, " size neg .75 mul dup rlto\n");
         fprintf(m_PSFileStream, " %s stroke} def\n\n", m_MarkerStrokeGrayColor);
         break;
         
      default:
         break;
   }
}

/*********************************************************************************/

void PS_Pointer (FLOAT X, FLOAT Y, PS_POINTER_TYPE Type, FLOAT Size)
{
   fprintf(m_PSFileStream, "\ngs newpath nd\n");
   fprintf(m_PSFileStream, "/size %g def\n", Size);
   fprintf(m_PSFileStream, "/x %g def\t/y %g def\n", X, Y);

   switch(Type)
   {
      case PS_POINTER_RIGHT:
         fprintf(m_PSFileStream,"\tsize 8 div slw\tx xs y ys mto\n");
         fprintf(m_PSFileStream,"\tsize neg size 3. div rlto\t0 size 1.5 div neg rlto\n");
         fprintf(m_PSFileStream,"\tcp gs %s fill gr %s stroke\n", m_MarkerFillGrayColor, m_MarkerStrokeGrayColor);
         break;

      case PS_POINTER_DOWN:
         fprintf(m_PSFileStream,"\tsize 8 div slw\tx xs y ys mto\n");
         fprintf(m_PSFileStream,"\tsize 3. div size rlto\tsize 1.5 div neg 0 rlto\n");
         fprintf(m_PSFileStream,"\tcp gs %s fill gr %s stroke\n", m_MarkerFillGrayColor, m_MarkerStrokeGrayColor);
         break;

      case PS_POINTER_LEFT:
         fprintf(m_PSFileStream, "\tsize 8 div slw\tx xs y ys mto\n");
         fprintf(m_PSFileStream, "\tsize size 3. div rlto\t0 size 1.5 div neg rlto\n");
         fprintf(m_PSFileStream, "\tcp gs %s fill gr %s stroke\n", m_MarkerFillGrayColor, m_MarkerStrokeGrayColor);
         break;
         
      case PS_POINTER_UP:
         fprintf(m_PSFileStream, "\tsize 8 div slw\tx xs y ys mto\n");
         fprintf(m_PSFileStream, "\tsize 3. div size neg rlto\tsize 1.5 div neg 0 rlto\n");
         fprintf(m_PSFileStream, "\tcp gs %s fill gr %s stroke\n", m_MarkerFillGrayColor, m_MarkerStrokeGrayColor);
         break;
   }
   fprintf(m_PSFileStream," gr\n");
}

/*********************************************************************************/

void PS_Mesh (void)
{
   fprintf(m_PSFileStream, "\n\n%% Mesh for plotting domain.\n");
   fprintf(m_PSFileStream, "savlw\t0.2 slw [1 2] 0 setdash\n");

   if (m_XAxisLog == TRUE)
   {
      PS_LogXMesh ();
   }
   else 
   {
      PS_XMesh ();
   }
   
   if (m_YAxisLog == TRUE) 
   {
      PS_LogYMesh ();
   }
   else 
   {
      PS_YMesh ();
   }
   
   fprintf(m_PSFileStream, "clw slw nd\n");
   fprintf(m_PSFileStream, "%% END Mesh for plotting domain.\n");
}

/*********************************************************************************/

void PS_XMesh (void)
{
   fprintf(m_PSFileStream, "/x {%g cvr} def\n", m_XMin);
   fprintf(m_PSFileStream, "x xtix %g cvr {xs dup 0 mto\n", m_XMax);
   fprintf(m_PSFileStream, "\tysize lto\n");
   fprintf(m_PSFileStream, "\tstroke xtix x add} for\n");
}

/*********************************************************************************/

void PS_YMesh (void)
{
   fprintf(m_PSFileStream, "/y {%g cvr} def\n",m_YMin);
   fprintf(m_PSFileStream, "y ytix %g cvr {ys dup 0 exch mto\n", m_YMax);
   fprintf(m_PSFileStream, "\txsize exch lto\n");
   fprintf(m_PSFileStream, "\tstroke ytix y add} for\n");
}

/*********************************************************************************/

void PS_LogXMesh (void)
{
   DOUBLE c;
   SHORT intmin;
   SHORT intdec;
   USHORT mask;
   UBYTE i;

   intmin = m_XMin;
   intdec = m_LowestDecadeX / 10;
   if (m_LogTicSelectMask < PS_MAXBIT) 
   {
      mask = PS_MINBIT;
      if((mask & m_LogTicSelectMask) != 0)
      {
         PS_XMinMeshLineSet(m_XMin);
      }
      mask = PS_MAXBIT;
      if((mask & m_LogTicSelectMask) != 0)
      {
         PS_XMeshLineSet(m_XMax);
      }
      
      if (intmin != intdec)
      {
         PS_XMeshLineSet(m_LowestDecadeX);
      }
      
      mask = PS_MINBIT;
      for(i = 1; i < 10; i++) 
      {
         mask = mask << 1;
         if((mask & m_LogTicSelectMask) != 0)
         {
            PS_XMeshLineSet(PS_XBottom(i));
         }
      }
   }
   else 
   {
      fprintf(m_PSFileStream, "/x %g cvr def\n", m_XMin);
      fprintf(m_PSFileStream, "x %g cvr %g cvr {xs dup 0 mto\n", (m_LowestDecadeX / 10), m_LowestDecadeX);
      fprintf(m_PSFileStream, "\tysize lto\n");
      fprintf(m_PSFileStream, "\tstroke} for\n");

      fprintf(m_PSFileStream, "/x %g cvr def\n", m_UpperDecadeX);
      fprintf(m_PSFileStream, "x %g cvr %g cvr {xs dup 0 mto\n", m_UpperDecadeX, m_XMax);
      fprintf(m_PSFileStream, "\tysize lto\n");
      fprintf(m_PSFileStream, "\tstroke} for\n");

      if (m_UpperDecadeX > m_LowestDecadeX) 
      {
         for (c = (10 * m_LowestDecadeX); c <= m_XMax; c *= 10)
         {
            fprintf(m_PSFileStream, "/x {%g cvr} def\n",c);
            fprintf(m_PSFileStream, "x 10 div dup x {xs dup 0 mto\n");
            fprintf(m_PSFileStream, "\tysize lto\n");
            fprintf(m_PSFileStream, "\tstroke} for\n");
         }
      }
   }
}

/*********************************************************************************/

void PS_XMeshLineSet (FLOAT Min)
{
   FLOAT x;
   
   for (x = Min; x <= m_XMax; x *= 10) 
   {
      fprintf(m_PSFileStream, "%g cvr xs 0 mto\n", x);
      fprintf(m_PSFileStream, "0 ysize rlto\n");
      fprintf(m_PSFileStream, "stroke\n");
   }
}

/*********************************************************************************/

void PS_XMinMeshLineSet (FLOAT Min)
{
   fprintf(m_PSFileStream, "%g cvr xs 0 mto\n", Min);
   fprintf(m_PSFileStream, "0 ysize rlto\n");
   fprintf(m_PSFileStream, "stroke\n");
}

/*********************************************************************************/

void PS_LogYMesh(void)
{
   DOUBLE c;
   USHORT mask;
   UBYTE i;

   if (m_LogTicSelectMask < PS_MAXBIT)
   {
      mask = PS_MINBIT;
      if((mask & m_LogTicSelectMask) != 0)
      {
         PS_YMinMeshLineSet(m_YMin);
      }
      mask = PS_MAXBIT;
      if((mask & m_LogTicSelectMask) != 0)
      {
         PS_YMeshLineSet(m_YMax);
      }
      if (((SHORT) m_YMin) != ((SHORT) (m_LowestDecadeY / 10)))
      {
         PS_YMeshLineSet(m_LowestDecadeY);
		}
      
      mask = PS_MINBIT;
      for(i = 1; i < 10; i++) 
      {
         mask = mask << 1;
         if((mask & m_LogTicSelectMask) != 0)
         {
            PS_YMeshLineSet(PS_YBottom(i));
         }
      }
   }
   else 
   {
      fprintf(m_PSFileStream, "/y %g cvr def\n",m_YMin);
      fprintf(m_PSFileStream, "y %g cvr %g cvr {ys dup 0 exch mto\n", (m_LowestDecadeY / 10), m_LowestDecadeY);
      fprintf(m_PSFileStream, "\txsize exch lto\n");
      fprintf(m_PSFileStream, "\tstroke} for\n");

      fprintf(m_PSFileStream, "/y %g cvr def\n", m_UpperDecadeY);
      fprintf(m_PSFileStream, "y %g cvr %g cvr {ys dup 0 exch mto\n", m_UpperDecadeY, m_YMax);
      fprintf(m_PSFileStream, "\txsize exch lto\n");
      fprintf(m_PSFileStream, "\tstroke} for\n");

      if (m_UpperDecadeY > m_LowestDecadeY) 
      {  
         for (c = (10 * m_LowestDecadeY); c <= m_UpperDecadeY; c *= 10)
         {   
            fprintf(m_PSFileStream, "/y {%g cvr} def\n", c);
            fprintf(m_PSFileStream, "y 10 div dup y {ys dup 0 exch mto\n");
            fprintf(m_PSFileStream, "\txsize exch lto\n");
            fprintf(m_PSFileStream, "\tstroke } for\n");
         }       
      }
   }
}

/*********************************************************************************/

void PS_YMeshLineSet (FLOAT Min)
{
   FLOAT y;
   
   for (y = Min; y <= m_YMax; y *= 10) 
   {
      fprintf(m_PSFileStream, "0 %g cvr ys mto\n", y);
      fprintf(m_PSFileStream, "xsize 0 rlto\n");
      fprintf(m_PSFileStream, "stroke\n");
   }
}

/*********************************************************************************/

void PS_YMinMeshLineSet(FLOAT Min)
{
   fprintf(m_PSFileStream, "0 %g cvr ys mto\n", Min);
   fprintf(m_PSFileStream, "xsize 0 rlto\n");
   fprintf(m_PSFileStream, "stroke\n");
}

/*********************************************************************************/

void PS_Move (FLOAT XPos, FLOAT YPos)
{
	if (m_MarkerMovementUnits == PS_DOMAIN)
   {
      fprintf(m_PSFileStream, "%g %g sm\n", XPos, YPos);
	}
   else
   {
      fprintf(m_PSFileStream, "%g %g mto\n", (XPos * m_Scale), (YPos * m_Scale));
   }
}

/*********************************************************************************/

void PS_Line (FLOAT XPos, FLOAT YPos)
{
	if (m_MarkerMovementUnits == PS_DOMAIN)
   {
      fprintf(m_PSFileStream, "%g %g sl\n", XPos, YPos);
	}
   else
   {
      fprintf(m_PSFileStream, "%g %g lto\n", (XPos * m_Scale), (YPos * m_Scale));
   }
}

/*********************************************************************************/

void PS_RectFill (FLOAT XPos, FLOAT YPos, FLOAT Width, FLOAT Height)
{
	PS_Move (XPos, YPos);
	PS_Line (XPos + Width, YPos);
	PS_Line (XPos + Width, YPos + Height);
	PS_Line (XPos, YPos + Height);
	PS_ClosePath();
	PS_Fill();
}

/*********************************************************************************/

void PS_ShiftOriginRelative (FLOAT XPos, FLOAT YPos)
{
   fprintf(m_PSFileStream, "\n\n%% Origin shift relative.\n");
   fprintf(m_PSFileStream, "/xs {%g mul} def /ys {%g mul} def\n", m_Scale, m_Scale);
   fprintf(m_PSFileStream, "%g %g mto  ", (XPos * m_Scale), (YPos * m_Scale));
   fprintf(m_PSFileStream, "cpt\n");
   fprintf(m_PSFileStream, "%% END Origin shift relative.\n");

   m_OriginX += (XPos * m_Scale);
   m_OriginY += (YPos * m_Scale);
}

/*********************************************************************************/

void PS_ShiftOriginAbsolute (FLOAT XPos, FLOAT YPos)
{
   fprintf(m_PSFileStream, "\n\n%% Origin shift absolute.\n");
   fprintf(m_PSFileStream, "/xs {%g mul} def /ys {%g mul} def\n", m_Scale, m_Scale);
   fprintf(m_PSFileStream, "%g %g mto  ", (XPos * m_Scale) - m_OriginX, (YPos * m_Scale) - m_OriginY);
   fprintf(m_PSFileStream, "cpt\n");
   fprintf(m_PSFileStream, "%% END Origin shift absolute.\n");

   m_OriginX = (XPos * m_Scale);
   m_OriginY = (YPos * m_Scale);
}

/*********************************************************************************/

void PS_Fill (void)
{
   fprintf(m_PSFileStream, " fill ");
}

/*********************************************************************************/

void PS_EveryOtherFill (void)
{
   fprintf(m_PSFileStream, " eofill ");
}

/*********************************************************************************/

void PS_ClosePath (void)
{
   fprintf(m_PSFileStream, " closepath ");
}

/*********************************************************************************/

void PS_GraphicsStateSave (void)
{
	fprintf(m_PSFileStream, " gs ");
}

/*********************************************************************************/

void PS_GraphicsStateRestore (void)
{
	fprintf(m_PSFileStream, " gr ");
}

/*********************************************************************************/

void PS_InsertPostScriptCommand (BYTE *PostScriptCommand) 
{
   if (PostScriptCommand != NULL)
	{
      fprintf(m_PSFileStream, "%s", PostScriptCommand);
   }
}

/*********************************************************************************/

void PS_MoveRelative (FLOAT XPos, FLOAT YPos)
{
	if (m_MarkerMovementUnits == PS_DOMAIN)
   {
      fprintf(m_PSFileStream, "%g cxs %g cys rmto\n", XPos, YPos);
	}
   else
   {
      fprintf(m_PSFileStream, "%g cvr %g cvr rmto\n", (XPos * m_Scale), (YPos * m_Scale));
   }
}

/*********************************************************************************/

void PS_LineRelative (FLOAT XPos, FLOAT YPos)
{
	if (m_MarkerMovementUnits == PS_DOMAIN)
   {
      fprintf(m_PSFileStream, "%g cxs %g cys rlto\n", XPos, YPos);
	}
   else
   {
      fprintf(m_PSFileStream, "%g cvr %g cvr rlto\n", (XPos * m_Scale), (YPos * m_Scale));
   }
}

/*********************************************************************************/

void PS_Reset (void)
{
   PS_ShiftOriginAbsolute (-m_OriginX, -m_OriginY);
   PS_LineWidth (1.0);
   PS_SetGray (0.0);
   PS_SetDashPattern (PS_LINE_NO_DASH, 0);
}

/*********************************************************************************/

void PS_Stroke (void)
{
   fprintf(m_PSFileStream, " stroke\tnd\n");
}

/*********************************************************************************/

void PS_AxisEnable (BOOLEAN AxisEnable, BOOLEAN AxisNumberEnable)
{
	m_AxisEnable = AxisEnable;
	m_AxisNumberEnable = AxisNumberEnable;
}

/*********************************************************************************/

void PS_LogAxisStyle (USHORT NumberSelectMask, 
                        USHORT TicSelectMask,
                        UBYTE NumberDistanceFromTic,
                        PS_AXIS_TIC_AND_NUMBER_TYPE TicNumberPosition,
                        UBYTE TicLength,
                        UBYTE TicLength10)
{
   m_LogNumberSelectMask = (NumberSelectMask & 0x03FF);
   m_LogTicSelectMask = (TicSelectMask & 0x03FF);

   if (NumberDistanceFromTic <= 100)
   {
      m_LogNumberDistanceFromTic = NumberDistanceFromTic;
   }   
   
   m_LogTicNumberPosition = TicNumberPosition;
   
   if ((TicLength > 0) && (TicLength < 100))
   {
      m_LogTicLength = TicLength;
   }
   
   if ((TicLength10 > 0) && (TicLength10 < 100))
   {
      m_LogTicLength10 = TicLength10;
   }
   else 
   {
      m_LogTicLength10 = m_LogTicLength;
   }
}

/*********************************************************************************/

void PS_LinearAxisStyle (UBYTE Precision, 
                           UBYTE TicOffsetNumbering, 
                           UBYTE NumberDistanceFromTic, 
                           PS_AXIS_TIC_AND_NUMBER_TYPE TicNumberPosition, 
                           UBYTE TicLength)
{
   if (Precision < 11) 
   {
      m_Precision = Precision;
   }
   
   if (TicOffsetNumbering < 20) 
   {
      m_LinearTicOffsetNumbering = TicOffsetNumbering;
   }
   
   if (NumberDistanceFromTic <= 100)
   {
      m_LinearNumberDistanceFromTic = NumberDistanceFromTic;
   }
   
   m_LinearTicNumberPosition = TicNumberPosition;
   
   if ((TicLength > 0) && (TicLength < 100))
   {
      m_LinearTicLength = TicLength;
   }
}

/*********************************************************************************/

void PS_Text (FLOAT XPos, FLOAT YPos, FLOAT Direction, BYTE *Text)
{
   BYTE FixedText[PS_MAX_STRING_LENGTH];
   
   fprintf(m_PSFileStream, "\n%% Text line.\n");
   if (m_TextMovementUnits == PS_DOMAIN)
   {
      fprintf(m_PSFileStream, "%g %g sm\n", XPos, YPos);
   }
   else
   {		
      fprintf(m_PSFileStream, "%g %g mto\n", (XPos * m_Scale), (YPos * m_Scale));
   }

   PS_FixString (Text, FixedText, TRUE);
   if (Direction == 0)
   {
      fprintf(m_PSFileStream, "(%s) show\n", FixedText);
	}
   else 
   {
      fprintf(m_PSFileStream, " cpt\t%.2f rotate\t(%s) show\n", Direction, FixedText);
      fprintf(m_PSFileStream, "0 0 mto %.2f rotate\t", -Direction);
      if (m_TextMovementUnits == PS_DOMAIN)
      {
         fprintf(m_PSFileStream, "%g ys -1.0 mul %g ys -1.0 mul ", XPos, YPos);
      }
      else
      {		
         fprintf(m_PSFileStream, "%g %g ", -(XPos * m_Scale), -(YPos * m_Scale));
      }
      fprintf(m_PSFileStream, " mto cpt\n");
   }
   fprintf(m_PSFileStream, "%% END Text line.\n");
}

/*********************************************************************************/

void PS_FixString (BYTE *Text, BYTE *FixedText, BOOLEAN FixMinus)
{
   UBYTE i = 0;
   UBYTE j = 0;
	
   while(Text[i] != '\0')
   {
      if(Text[i] == '-')
      {
         if(FixMinus == TRUE)
         {
            /*
               Substitute with octal 261 
            */
            FixedText[j++] = '\\';
            FixedText[j++] = '2';
            FixedText[j++] = '6';
            FixedText[j++] = '1';
         }
      }
      else if((Text[i] == '\\') || (Text[i] == '(') || (Text[i] == ')'))
      {
         FixedText[j++] = '\\';
         FixedText[j++] = Text[i];
      }
      else
		{
         FixedText[j++] = Text[i];
      }
      i++;
   }
   FixedText[j] = '\0';	
}

/*********************************************************************************/

void PS_LineWidth (FLOAT LineWidth)
{
	fprintf(m_PSFileStream, "%8.5f slw\n", LineWidth);
}

/*********************************************************************************/

void PS_XAxis (FLOAT XSize, FLOAT XMin, FLOAT XMax, FLOAT Offset, FLOAT XTicSeparation, UBYTE NumberSeparation)
{
   FLOAT x;
   FLOAT XRange;
   FLOAT XScale;
   SHORT NumberDistanceFromTic;
   BYTE FixedText[PS_MAX_STRING_LENGTH];
   SHORT TextTop = 0;
   SHORT TextBottom = 0;
   BYTE Line[PS_MAX_STRING_LENGTH];
   UBYTE SignificantDecimal;

   m_XAxisLog = FALSE;
   m_RealXMax = XSize;
   m_XMax = XMax + (0.001 * (XMax - XMin));
   m_XMin = XMin;
   XRange = m_XMax - m_XMin;
   XScale = (XSize * m_Scale) / XRange;
   NumberDistanceFromTic = (m_LinearNumberDistanceFromTic * (m_FontSize / 10.0)) + 1;

   fprintf(m_PSFileStream, "\n\n%% X axis (linear).\n");
   fprintf(m_PSFileStream, "/xs {%g cvr mul} def\n", XScale); 
   fprintf(m_PSFileStream, "/xtix {%g cvr} def\n", XTicSeparation); 
   fprintf(m_PSFileStream, "/xsize {%g cvr} def\n", (XSize * m_Scale)); 
   fprintf(m_PSFileStream, "/xrange {%g cvr} def\n", XRange);
   fprintf(m_PSFileStream, "/xos {%g cvr sub} def\n\n", m_XMin);

   if(m_AxisEnable == TRUE) 
   {
      if ((m_LinearTicNumberPosition == PS_AXIS_TIC_AND_NUMBER_TIC_RIGHT_ABOVE_NUM_LEFT_BELOW) || 
          (m_LinearTicNumberPosition == PS_AXIS_TIC_AND_NUMBER_TIC_RIGHT_ABOVE_NUM_RIGHT_ABOVE))
      {
         TextBottom = 0;
      }
      else
      { 
         TextBottom = -m_LinearTicLength;
      }
	
      if ((m_LinearTicNumberPosition == PS_AXIS_TIC_AND_NUMBER_TIC_LEFT_BELOW_NUM_LEFT_BELOW) || 
          (m_LinearTicNumberPosition == PS_AXIS_TIC_AND_NUMBER_TIC_LEFT_BELOW_NUM_RIGHT_ABOVE))
      {
         TextTop = 0;
      }
      else
      { 
         TextTop = m_LinearTicLength;
      }
      
      fprintf(m_PSFileStream, "/x 0 def\n");
      fprintf(m_PSFileStream, "0 %g mto\n%g 0 rlto\nstroke\n", (Offset * m_Scale), (XSize * m_Scale));
      fprintf(m_PSFileStream, "x 0 xtix xrange {xs dup %g mto\n", (Offset * m_Scale) + TextTop);
      fprintf(m_PSFileStream, "\t%g lto\n", (Offset * m_Scale) + TextBottom);
      fprintf(m_PSFileStream, "\tstroke xtix x add} for\n");
   }

   if(m_AxisNumberEnable == TRUE)
   {
      SignificantDecimal = PS_FindSignificantDecimalPoint(m_XMin, (XRange * 1.01), NumberSeparation, XTicSeparation);
    
      for (x = (m_LinearTicOffsetNumbering * XTicSeparation); x <= (XRange * 1.01); x += (NumberSeparation * XTicSeparation))
      {
         PS_FormAxisNumber(x + m_XMin, SignificantDecimal, Line);
         fprintf(m_PSFileStream, "\nnewpath 1000 1000 mto\n");
         fprintf(m_PSFileStream, "(%s) stw pop 2 div\n", Line);
         fprintf(m_PSFileStream, "%g cvr exch sub ", (x * XScale)); 

         if ((m_LinearTicNumberPosition == PS_AXIS_TIC_AND_NUMBER_TIC_LEFT_BELOW_NUM_LEFT_BELOW) || 
             (m_LinearTicNumberPosition == PS_AXIS_TIC_AND_NUMBER_TIC_THROUGH_NUM_LEFT_BELOW) ||
             (m_LinearTicNumberPosition == PS_AXIS_TIC_AND_NUMBER_TIC_RIGHT_ABOVE_NUM_LEFT_BELOW))
         {
            fprintf(m_PSFileStream, "%g charheight sub ", (Offset * m_Scale) + TextBottom - NumberDistanceFromTic);
         }
         else
         {
            fprintf(m_PSFileStream, "%g ", (Offset * m_Scale) + TextTop + NumberDistanceFromTic);
         }
         
         PS_FixString (Line, FixedText, TRUE);
         fprintf(m_PSFileStream, "mto (%s) show\n", FixedText);
      }
   }

   m_YPositionOfXAxis = (Offset * m_Scale) + TextBottom - NumberDistanceFromTic - ((1.4) * m_FontSize);
   fprintf(m_PSFileStream, "\n/xs {%g cvr sub %g cvr mul} def\n", m_XMin, XScale); 
   fprintf(m_PSFileStream, "%% END X axis (linear).\n");
}

/*********************************************************************************/

void PS_YAxis (FLOAT YSize, FLOAT YMin, FLOAT YMax, FLOAT Offset, FLOAT YTicSeparation, UBYTE NumberSeparation)
{
   FLOAT y;
   FLOAT YRange;
   FLOAT YScale;
   SHORT NumberDistanceFromTic;
   BYTE FixedText[PS_MAX_STRING_LENGTH];
   SHORT TextRight = 0;
   SHORT TextLeft = 0;
   BYTE Line[PS_MAX_STRING_LENGTH];
   UBYTE SignificantDecimal;

   m_YAxisLog = FALSE;
   m_RealYMax = YSize;
   m_YMax = YMax + (0.001 * (YMax - YMin));
   m_YMin = YMin;
   YRange = m_YMax - m_YMin;
   YScale = (YSize * m_Scale) / YRange;
   NumberDistanceFromTic = m_LinearNumberDistanceFromTic + 1;

   fprintf(m_PSFileStream, "\n\n%% Y axis (linear).\n");
   fprintf(m_PSFileStream, "/ys {%g cvr mul} def\n", YScale);
   fprintf(m_PSFileStream, "/ytix {%g cvr} def\n", YTicSeparation);
   fprintf(m_PSFileStream, "/ysize {%g cvr} def\n", (YSize * m_Scale)); 
   fprintf(m_PSFileStream, "/yrange {%g cvr} def\n", (YRange * 1.01));
   fprintf(m_PSFileStream, "/yos {%g cvr sub} def\n", m_YMin);

   if(m_AxisEnable == TRUE)
   {
      fprintf(m_PSFileStream, "%g 0 mto\n0 %g rlto\nstroke\n", (Offset * m_Scale), (YSize * m_Scale));
      if ((m_LinearTicNumberPosition == PS_AXIS_TIC_AND_NUMBER_TIC_RIGHT_ABOVE_NUM_LEFT_BELOW) || 
          (m_LinearTicNumberPosition == PS_AXIS_TIC_AND_NUMBER_TIC_RIGHT_ABOVE_NUM_RIGHT_ABOVE))
      {
         TextLeft = 0;
      }
      else
      { 
         TextLeft = -m_LinearTicLength;
      }
      
      if ((m_LinearTicNumberPosition == PS_AXIS_TIC_AND_NUMBER_TIC_LEFT_BELOW_NUM_LEFT_BELOW) || 
          (m_LinearTicNumberPosition == PS_AXIS_TIC_AND_NUMBER_TIC_LEFT_BELOW_NUM_RIGHT_ABOVE))
      {
         TextRight = 0;
      }
      else
      { 
         TextRight = m_LinearTicLength;
      }
      
      fprintf(m_PSFileStream, "/y 0 def\n");
      fprintf(m_PSFileStream, "y 0 ytix yrange {ys dup %g exch mto\n", (Offset * m_Scale) + TextRight);
      fprintf(m_PSFileStream, "\t%g exch lto\n", (Offset * m_Scale) + TextLeft);
      fprintf(m_PSFileStream, "\tstroke ytix y add} for\n");
   }

   if(m_AxisNumberEnable == TRUE) 
   {
      SignificantDecimal = PS_FindSignificantDecimalPoint(m_YMin, (YRange * 1.01), NumberSeparation, YTicSeparation);

      for (y = (m_LinearTicOffsetNumbering * YTicSeparation); y <= (YRange * 1.01); y += (NumberSeparation * YTicSeparation)) 
      {
         PS_FormAxisNumber(y + m_YMin, SignificantDecimal, Line);
         fprintf(m_PSFileStream, "newpath 1000 1000 mto\n");
         if ((m_LinearTicNumberPosition == PS_AXIS_TIC_AND_NUMBER_TIC_LEFT_BELOW_NUM_LEFT_BELOW) || 
             (m_LinearTicNumberPosition == PS_AXIS_TIC_AND_NUMBER_TIC_THROUGH_NUM_LEFT_BELOW) ||
             (m_LinearTicNumberPosition == PS_AXIS_TIC_AND_NUMBER_TIC_RIGHT_ABOVE_NUM_LEFT_BELOW))
         {
            fprintf(m_PSFileStream, "%g (%s) stw pop sub\n", (Offset * m_Scale) + TextLeft - NumberDistanceFromTic, Line);
            m_XPositionOfYAxis = (Offset * m_Scale) + TextLeft - NumberDistanceFromTic;
         }
         else
         {
            fprintf(m_PSFileStream, "%g ", (Offset * m_Scale) + TextRight + NumberDistanceFromTic);
            m_XPositionOfYAxis = (Offset * m_Scale) + TextRight + NumberDistanceFromTic;
         }
	    
         fprintf(m_PSFileStream, "%g charheight ", YScale * y);
         fprintf(m_PSFileStream, "2 div sub mto ");

         PS_FixString (Line, FixedText, TRUE);
         fprintf(m_PSFileStream, "(%s) show\n", FixedText);
         if (strlen(FixedText) > strlen(m_StringAtHighestY))
         {
            sprintf(m_StringAtHighestY, "%s", FixedText);
         }
      }
   }

   fprintf(m_PSFileStream, "\n/ys {%g cvr sub %g cvr mul} def\n", YMin, YScale);
   fprintf(m_PSFileStream, "%% END Y axis (linear).\n");
}

/*********************************************************************************/

UBYTE PS_FindSignificantDecimalPoint (FLOAT Min, FLOAT Range, UBYTE NumberSeparation, FLOAT TickSeparation)
{
   FLOAT var;
   UBYTE i;
   UBYTE DecimalPointIndex;
   BOOLEAN ExponentMarker;
   UBYTE SignificantDecimal = 0;
   BYTE Line[PS_MAX_STRING_LENGTH];
	
   for (var = (m_LinearTicOffsetNumbering * TickSeparation); var <= Range; var += (NumberSeparation * TickSeparation)) 
   {
      if((var + Min) < pow(10.0, (FLOAT) m_Precision))
      {
         sprintf(Line, "%*g", m_Precision, (var + Min));
      }
      else 
      {
         sprintf(Line, "%*e", m_Precision, (var + Min));
      }

      ExponentMarker = FALSE;
      DecimalPointIndex = 0;
      i = 0;
      while (Line[i] != '\0') 
      {
         if (Line[i] == 'e') 
         {
            ExponentMarker = TRUE;
         }
         
         if (Line[i] == '.')
         {
            DecimalPointIndex = i;
         }
         i++;
      }
      
      if (DecimalPointIndex == 0)
      {
         DecimalPointIndex = i - 1;
      }
      
      if (ExponentMarker == FALSE)
      {
         SignificantDecimal = MAX((i - DecimalPointIndex - 1),SignificantDecimal);
      }
   }

   return(SignificantDecimal);
}

/*********************************************************************************/

void PS_FormAxisNumber (FLOAT Value, UBYTE SignificantDecimal, BYTE *Line)
{
   UBYTE i = 0;
   UBYTE DecimalPointIndex = 0;
   BOOLEAN ExponentMarker = FALSE;
   UBYTE ExponentIndex = 0;

   sprintf(Line, "%#*f", m_Precision, Value);

   while (Line[i] != '\0') 
   {
      if (Line[i] == 'e') 
      {
         ExponentMarker = TRUE;
         ExponentIndex = i;
      }
      
      if (Line[i] == '.')
      {
         DecimalPointIndex = i;
      }
      i++;
   }
   
   if (ExponentMarker == TRUE) 
   {
      PS_CleanExponentialString (Line, ExponentIndex);
      i = 0;
      while (Line[i] != '\0')
      {
         i++;
      }
   }
   
   if (DecimalPointIndex > 0) 
   {
      if (SignificantDecimal > 0)
      { 
         Line[SignificantDecimal + DecimalPointIndex + 1] = '\0';
      }
      else
      { 
         Line[DecimalPointIndex] = '\0';
      }
   }
}

/*********************************************************************************/

void PS_CleanExponentialString (BYTE *Line, UBYTE ExponentIndex)
{
   int i;
   
   if (Line[ExponentIndex - 1] == '.')
   {
      i = ExponentIndex - 1;
      ExponentIndex--;
      while (Line[i] != '\0')
      {
         Line[i] = Line[i + 1];
         i++; 
      }
   }
   
   if (Line[ExponentIndex + 1] == '+')
   {
      i = ExponentIndex + 1;
      ExponentIndex--;
      while (Line[i] != '\0')
      {
         Line[i] = Line[i + 1];
         i++; 
      }
   }
   
   while (Line[ExponentIndex + 2] == '0')
   {	
      i = ExponentIndex + 2;
      while (Line[i] != '\0')
      {
         Line[i] = Line[i + 1];
         i++;
      }
   }
}

/*********************************************************************************/

void PS_XLog (FLOAT XSize, FLOAT XMin, FLOAT XMax, FLOAT Offset)
{
   DOUBLE XRange;
   DOUBLE XScale;
   SHORT NumberDistanceFromTic;
   SHORT TextTop;
   SHORT TextBottom;
   SHORT TextTop10;
   SHORT TextBottom10;   

   m_XAxisLog = TRUE;
   m_RealXMax = XSize;
   m_XMax = PS_Trunc((XMax * 0.99), PS_ROUND_UP);
   m_XMin = PS_Trunc((XMin * 1.01), PS_ROUND_DOWN);
   m_LowestDecadeX = pow(10.0, ceil(log10(m_XMin) - 0.0001));
   m_UpperDecadeX = pow(10.0, floor(log10(m_XMax) + 0.0001));
   XRange = log10(m_XMax / m_XMin);
   XScale = (XSize * m_Scale) / XRange;
   NumberDistanceFromTic = (m_LogNumberDistanceFromTic * (m_FontSize / 10.0)) + 1;

   fprintf(m_PSFileStream, "\n\n%% X axis (log).\n");
   fprintf(m_PSFileStream, "/xs {log %.5g cvr sub %.5g cvr mul} def\n", log10(m_XMin), XScale); 
   fprintf(m_PSFileStream, "/xsize {%.5g cvr} def\n", (XSize * m_Scale)); 
   fprintf(m_PSFileStream, "/xrange {%.5g cvr} def\n", XRange);
   fprintf(m_PSFileStream, "/xos {%.5g cvr sub} def\n", (Offset * m_Scale));

   if(m_AxisEnable == TRUE) 
   {
      fprintf(m_PSFileStream, "0 %.5g mto\n%.5g 0 rlto\nstroke\n", (Offset * m_Scale), (XSize * m_Scale));
   }
   
   if ((m_LogTicNumberPosition == PS_AXIS_TIC_AND_NUMBER_TIC_RIGHT_ABOVE_NUM_LEFT_BELOW) || 
       (m_LogTicNumberPosition == PS_AXIS_TIC_AND_NUMBER_TIC_RIGHT_ABOVE_NUM_RIGHT_ABOVE))
   {
      TextBottom = 0;
      TextBottom10 = 0;
   }
   else
   {
      TextBottom = -m_LogTicLength;
      TextBottom10 = -m_LogTicLength10;
   }

   if ((m_LogTicNumberPosition == PS_AXIS_TIC_AND_NUMBER_TIC_LEFT_BELOW_NUM_LEFT_BELOW) || 
       (m_LogTicNumberPosition == PS_AXIS_TIC_AND_NUMBER_TIC_LEFT_BELOW_NUM_RIGHT_ABOVE))
   {
      TextTop = 0;
      TextTop10 = 0;
   }
   else
   {
      TextTop = m_LogTicLength;
      TextTop10 = m_LogTicLength10;
   }
   
   m_YPositionOfXAxis = (Offset * m_Scale) + TextBottom - NumberDistanceFromTic - ((1.4) * m_FontSize);
   PS_XLogTicNumbering ((Offset * m_Scale), TextTop, TextBottom, NumberDistanceFromTic, TextTop10, TextBottom10, XScale);
   fprintf(m_PSFileStream, "%% END X axis (log).\n");
}

/*********************************************************************************/

FLOAT PS_Trunc(FLOAT Value, PS_ROUND_TYPE Round)
{
   DOUBLE Log10Value;
   FLOAT Decade;
   FLOAT Result;

   Log10Value = floor(log10(Value));
   Decade = pow(10.0, Log10Value);
   Result = Value / Decade;
   
   switch (Round)
   {
      case PS_ROUND_DOWN:
         Result = floor(Result) * Decade;
         break;
         
      case PS_ROUND_UP:
         Result = ceil(Result) * Decade;
         break;
         
      case PS_ROUND_EVEN:
         Result = floor(Result + 0.5) * Decade;
         break;
   }

   return (Result);
}

/*********************************************************************************/

void PS_XLogTicNumbering (FLOAT Offset, 
                           SHORT TextTop, 
                           SHORT TextBottom, 
                           SHORT NumberDistanceFromTic, 
                           SHORT TextTop10, 
                           SHORT TextBottom10,
                           DOUBLE XScale)
{
   USHORT mask;
   UBYTE i;

   if(m_AxisEnable == TRUE) 
   {
      if (m_LogTicSelectMask < PS_MAXBIT)
      {
         mask = PS_MINBIT;
         if((mask & m_LogTicSelectMask) != 0)
         {
            PS_XSetTics(m_XMin, Offset + TextTop, Offset + TextBottom);
         }
       
         mask = PS_MAXBIT;
         if((mask & m_LogTicSelectMask) != 0)
         {
            PS_XSetTics(m_XMax, Offset + TextTop, Offset + TextBottom);
         }
         
         if (((SHORT) m_XMin) != ((SHORT) (m_LowestDecadeX / 10)))
         {
            PS_XSetTics(m_LowestDecadeX, Offset + TextTop, Offset + TextBottom);
         }
         
         mask = PS_MINBIT;	
         for(i = 1; i < 10; i++)
         {
            mask = mask << 1;
            if((mask & m_LogTicSelectMask) != 0)
            {
               if(i == 1)
               {
                  PS_XSetTics(PS_XBottom(1), Offset + TextTop10, Offset + TextBottom10);
               }
               else    
               {
                  PS_XSetTics(PS_XBottom(i), Offset + TextTop, Offset + TextBottom);
               }
            }
         }
      }
      else
      {
         PS_AllXSetTics(Offset + TextBottom, Offset + TextTop);
      }
   }

   if(m_AxisNumberEnable == TRUE)
   {
      if (m_LogNumberSelectMask < PS_MAXBIT)
      {
         mask = PS_MINBIT;
         if ((mask & m_LogNumberSelectMask) != 0)
         {
            PS_XSetMinNumbers(m_XMin, NumberDistanceFromTic, Offset, TextTop, TextBottom, XScale);
         }

         mask = PS_MAXBIT; 
         if ((mask & m_LogNumberSelectMask) != 0) 
         {
            PS_XSetNumbers (m_XMax, NumberDistanceFromTic, Offset, TextTop, TextBottom, XScale);
         }
         
         mask = PS_MINBIT;
         for(i = 1; i < 10; i++) 
         {
            mask = mask << 1;
            if((mask & m_LogNumberSelectMask) != 0)
            {
               PS_XSetNumbers(PS_XBottom(i), NumberDistanceFromTic, Offset, TextTop, TextBottom, XScale);
            }
         }
      }
      else 
      {
         PS_XSetMinNumbers (m_XMin, NumberDistanceFromTic, Offset, TextTop, TextBottom, XScale);
         PS_XSetNumbers (m_XMax, NumberDistanceFromTic, Offset, TextTop, TextBottom, XScale);
         PS_XSetNumbers (PS_XBottom(1), NumberDistanceFromTic, Offset, TextTop, TextBottom, XScale);
      }
   }
}

/*********************************************************************************/

void PS_XSetTics (FLOAT Min, FLOAT Top, FLOAT Bottom)
{
   FLOAT x;

   for (x = Min; x <= m_XMax; x *= 10) 
   {
      fprintf(m_PSFileStream, "%.5g cvr xs %.5g mto\n", x, Bottom);
      fprintf(m_PSFileStream, "0 %.5g rlto\n", Top - Bottom);
      fprintf(m_PSFileStream, "stroke\n");
   }
}

/*********************************************************************************/

void PS_AllXSetTics (FLOAT Top, FLOAT Bottom)
{
   FLOAT c;

   fprintf(m_PSFileStream, "/x %.5g cvr def\n", m_XMin);
   fprintf(m_PSFileStream, "x %.5g cvr %.5g cvr {xs dup %.5g mto\n", (m_LowestDecadeX / 10), m_LowestDecadeX, Top);
   fprintf(m_PSFileStream, " %.5g lto\n", Bottom);
   fprintf(m_PSFileStream, " stroke} for\n");
   fprintf(m_PSFileStream, "/x %.5g cvr def\n", m_UpperDecadeX);
   fprintf(m_PSFileStream, "x %.5g cvr %.5g cvr {xs dup %.5g mto\n", m_UpperDecadeX, m_XMax, Top);
   fprintf(m_PSFileStream, " %.5g lto\n", Bottom);
   fprintf(m_PSFileStream, " stroke} for\n");
	
   if (m_UpperDecadeX > m_LowestDecadeX) 
   {
      for (c = (10 * m_LowestDecadeX); c <= m_UpperDecadeX; c *= 10) 
      {
         fprintf(m_PSFileStream, "/x %.7f def\n", c);
         fprintf(m_PSFileStream, "x 10 div dup x {xs dup %.5g mto\n", Top);
         fprintf(m_PSFileStream, " %.5g lto\n", Bottom);
         fprintf(m_PSFileStream, " stroke} for\n");
      }
   }
}

/*********************************************************************************/

void PS_XSetNumbers (FLOAT Start, 
                     SHORT NumberDistanceFromTic, 
                     FLOAT Offset, 
                     SHORT TextTop, 
                     SHORT TextBottom,
                     DOUBLE XScale)
{
   FLOAT x;

   for (x = Start; x <= m_XMax; x *= 10) 
   {
      x = PS_Trunc(x, PS_ROUND_EVEN);
      fprintf(m_PSFileStream, "(%.5g) stw pop 2 div\n",x);
      fprintf(m_PSFileStream, "%.5g exch sub ", log10(x / m_XMin) * XScale); 
      if ((m_LogTicNumberPosition == PS_AXIS_TIC_AND_NUMBER_TIC_LEFT_BELOW_NUM_LEFT_BELOW) || 
          (m_LogTicNumberPosition == PS_AXIS_TIC_AND_NUMBER_TIC_THROUGH_NUM_LEFT_BELOW) ||
          (m_LogTicNumberPosition == PS_AXIS_TIC_AND_NUMBER_TIC_RIGHT_ABOVE_NUM_LEFT_BELOW))
      {
         fprintf(m_PSFileStream, "%.5g charheight sub\n", Offset + TextBottom - NumberDistanceFromTic);
      }
      else 
      {
         fprintf(m_PSFileStream, "%.5g ", Offset + TextTop + NumberDistanceFromTic);
      }
      fprintf(m_PSFileStream, "mto (%.5g) show\n", x);
   }
}

/*********************************************************************************/

void PS_XSetMinNumbers (FLOAT Start, 
                        SHORT NumberDistanceFromTic, 
                        FLOAT Offset, 
                        SHORT TextTop, 
                        SHORT TextBottom,
                        DOUBLE XScale)
{
   FLOAT x;

	x = PS_Trunc(Start, PS_ROUND_EVEN);
   fprintf(m_PSFileStream, "(%.5g) stw pop 2 div\n", x);
   fprintf(m_PSFileStream, "%.5g exch sub ", log10(x / m_XMin) * XScale); 
   if ((m_LogTicNumberPosition == PS_AXIS_TIC_AND_NUMBER_TIC_LEFT_BELOW_NUM_LEFT_BELOW) || 
       (m_LogTicNumberPosition == PS_AXIS_TIC_AND_NUMBER_TIC_THROUGH_NUM_LEFT_BELOW) ||
       (m_LogTicNumberPosition == PS_AXIS_TIC_AND_NUMBER_TIC_RIGHT_ABOVE_NUM_LEFT_BELOW))
   {
      fprintf(m_PSFileStream, "%.5g charheight sub\n", Offset + TextBottom - NumberDistanceFromTic);
   }
   else
   {
      fprintf(m_PSFileStream, "%.5g ", Offset + TextTop + NumberDistanceFromTic);
   }
   fprintf(m_PSFileStream, "mto (%.5g) show\n", x);
}

/*********************************************************************************/

FLOAT PS_XBottom (FLOAT Multiplier)
{
   FLOAT Bottom;

   Bottom = m_LowestDecadeX * Multiplier / 10.0;
   while (Bottom < m_XMin) 
   {
      Bottom *= 10.0;
   }
   
   return(Bottom);
}

/*********************************************************************************/

FLOAT PS_YBottom (FLOAT Multiplier)
{
   FLOAT Bottom;

   Bottom = m_LowestDecadeY * Multiplier / 10.0;
   while (Bottom < m_YMin) 
   {
      Bottom *= 10.0;
   }
   
   return(Bottom);
}

/*********************************************************************************/

void PS_YLog (FLOAT YSize, FLOAT YMin, FLOAT YMax, FLOAT Offset)
{
   DOUBLE YRange;
   DOUBLE YScale;
   SHORT NumberDistanceFromTic;
   SHORT TextRight;
   SHORT TextLeft;
   SHORT TextRight10;
   SHORT TextLeft10;   

   m_YAxisLog = TRUE;
   m_RealYMax = YSize;
   m_YMax = PS_Trunc((YMax * 0.99), PS_ROUND_UP);
   m_YMin = PS_Trunc((YMin * 1.01), PS_ROUND_DOWN);
   m_LowestDecadeY = pow(10.0, ceil(log10(m_YMin) - 0.0001));
   m_UpperDecadeY = pow(10.0, floor(log10(m_YMax) + 0.0001));
   YRange = log10(m_YMax / m_YMin);
   YScale = (YSize * m_Scale) / YRange;
   NumberDistanceFromTic = (m_LogNumberDistanceFromTic * (m_FontSize / 10.0)) + 1;

   fprintf(m_PSFileStream, "\n\n%% Y axis (log).\n");
   fprintf(m_PSFileStream, "/ys {log %.5g cvr sub %.5g cvr mul} def\n", log10(m_YMin), YScale); 
   fprintf(m_PSFileStream, "/ysize {%.5g cvr} def\n", (YSize * m_Scale)); 
   fprintf(m_PSFileStream, "/yrange {%.5g cvr} def\n", YRange);
   fprintf(m_PSFileStream, "/yos {%.5g cvr sub} def\n", (Offset * m_Scale));

   if(m_AxisEnable == TRUE) 
   {
      fprintf(m_PSFileStream, "%.5g 0 mto\t0 %.5g rlto\tstroke\n", (Offset * m_Scale), (YSize * m_Scale));
   }
   
   if ((m_LogTicNumberPosition == PS_AXIS_TIC_AND_NUMBER_TIC_RIGHT_ABOVE_NUM_LEFT_BELOW) || 
       (m_LogTicNumberPosition == PS_AXIS_TIC_AND_NUMBER_TIC_RIGHT_ABOVE_NUM_RIGHT_ABOVE))
   {
      TextLeft = 0;
      TextLeft10 = 0;
   }
   else
   {
      TextLeft = -m_LogTicLength;
      TextLeft10 = -m_LogTicLength10;
   }

   if ((m_LogTicNumberPosition == PS_AXIS_TIC_AND_NUMBER_TIC_LEFT_BELOW_NUM_LEFT_BELOW) || 
       (m_LogTicNumberPosition == PS_AXIS_TIC_AND_NUMBER_TIC_LEFT_BELOW_NUM_RIGHT_ABOVE))
   {
      TextRight = 0;
      TextRight10 = 0;
   }
   else
   {
      TextRight = m_LogTicLength;
      TextRight10 = m_LogTicLength10;
   }
   
   m_XPositionOfYAxis = (Offset * m_Scale) + TextLeft - NumberDistanceFromTic;
   PS_YLogTicNumbering ((Offset * m_Scale), TextRight, TextLeft, NumberDistanceFromTic, TextRight10, TextLeft10, YScale);
   fprintf(m_PSFileStream, "%% END Y axis (log).\n");
}

/*********************************************************************************/

void PS_YLogTicNumbering (FLOAT Offset, 
                           SHORT TextRight, 
                           SHORT TextLeft, 
                           SHORT NumberDistanceFromTic, 
                           SHORT TextRight10, 
                           SHORT TextLeft10,
                           DOUBLE YScale)
{
   USHORT mask;
   UBYTE i;

   if(m_AxisEnable == TRUE) 
   {
      if (m_LogTicSelectMask < PS_MAXBIT)
      {
         mask = PS_MINBIT;
         if((mask & m_LogTicSelectMask) != 0)
         {
            PS_YSetMinTics(m_YMin, Offset + TextRight, Offset + TextLeft);
         }
       
         mask = PS_MAXBIT;
         if((mask & m_LogTicSelectMask) != 0)
         {
            PS_YSetTics(m_YMax, Offset + TextRight, Offset + TextLeft);
         }
         
         if (((SHORT) m_YMin) != ((SHORT) (m_LowestDecadeY / 10)))
         {
            PS_YSetTics(m_LowestDecadeY, Offset + TextRight, Offset + TextLeft);
         }
         
         mask = PS_MINBIT;	
         for(i = 1; i < 10; i++)
         {
            mask = mask << 1;
            if((mask & m_LogTicSelectMask) != 0)
            {
               if(i == 1)
               {
                  PS_YSetTics(PS_YBottom(1), Offset + TextRight10, Offset + TextLeft10);
               }
               else    
               {
                  PS_YSetTics(PS_YBottom(i), Offset + TextRight, Offset + TextLeft);
               }
            }
         }
      }
      else
      {
         PS_AllYSetTics(Offset + TextLeft, Offset + TextRight);
      }
   }

   if(m_AxisNumberEnable == TRUE)
   {
      if (m_LogNumberSelectMask < PS_MAXBIT)
      {
         mask = PS_MINBIT;
         if ((mask & m_LogNumberSelectMask) != 0)
         {
            PS_YSetMinNumbers(m_YMin, NumberDistanceFromTic, Offset, TextRight, TextLeft, YScale);
         }

         mask = PS_MAXBIT; 
         if ((mask & m_LogNumberSelectMask) != 0) 
         {
            PS_YSetNumbers (m_YMax, NumberDistanceFromTic, Offset, TextRight, TextLeft, YScale);
         }
         
         mask = PS_MINBIT;
         for(i = 1; i < 10; i++) 
         {
            mask = mask << 1;
            if((mask & m_LogNumberSelectMask) != 0)
            {
               PS_YSetNumbers(PS_YBottom(i), NumberDistanceFromTic, Offset, TextRight, TextLeft, YScale);
            }
         }
      }
      else 
      {
         PS_YSetMinNumbers(m_YMin, NumberDistanceFromTic, Offset, TextRight, TextLeft, YScale);
         PS_YSetNumbers (m_YMax, NumberDistanceFromTic, Offset, TextRight, TextLeft, YScale);
         PS_YSetNumbers(PS_YBottom(1), NumberDistanceFromTic, Offset, TextRight, TextLeft, YScale);
      }
   }
}

/*********************************************************************************/

void PS_YSetTics (FLOAT Min, FLOAT Right, FLOAT Left)
{
   FLOAT y;

   for (y = Min; y <= m_YMax; y *= 10) 
   {
      fprintf(m_PSFileStream, "%.5g %.5g cvr ys mto\n", Right, y);
      fprintf(m_PSFileStream, "%.5g 0 rlto\n", Left - Right);
      fprintf(m_PSFileStream, "stroke\n");
   }
}

/*********************************************************************************/

void PS_YSetMinTics (FLOAT Min, FLOAT Right, FLOAT Left)
{
   fprintf(m_PSFileStream, "%.5g %.5g cvr ys mto\n", Right, Min);
   fprintf(m_PSFileStream, "%.5g 0 rlto\n", Left - Right);
   fprintf(m_PSFileStream, "stroke\n");
}

/*********************************************************************************/

void PS_AllYSetTics (FLOAT Left, FLOAT Right)
{
   FLOAT c;

   fprintf(m_PSFileStream, "/y %g cvr def\n", m_YMin);
   fprintf(m_PSFileStream, "y %.5g cvr %.5g cvr {ys dup %.5g exch mto\n", (m_LowestDecadeY / 10), m_LowestDecadeY, Right);
   fprintf(m_PSFileStream, "\t%.5g exch lto\n", Left);
   fprintf(m_PSFileStream, "\tstroke} for\n");
   fprintf(m_PSFileStream, "/y %.5g cvr def\n", m_UpperDecadeY);
   fprintf(m_PSFileStream, "y %.5g cvr %.5g cvr {ys dup %.5g exch mto\n", m_UpperDecadeY, m_YMax, Right);
   fprintf(m_PSFileStream, "\t%.5g exch lto\n", Left);
   fprintf(m_PSFileStream, "\tstroke} for\n");
	
   if (m_UpperDecadeY > m_LowestDecadeY) 
   {
      for (c = (10 * m_LowestDecadeY); c <= m_UpperDecadeY; c *= 10) 
      {
         fprintf(m_PSFileStream, "/y %.5g def\n", c);
         fprintf(m_PSFileStream, "y 10 div dup y {ys dup %.5g exch mto\n", Right);
         fprintf(m_PSFileStream, "\t%.5g exch lto\n", Left);
         fprintf(m_PSFileStream, "\tstroke} for\n");
      }
   }
}

/*********************************************************************************/

void PS_YSetNumbers (FLOAT Start, 
                     SHORT NumberDistanceFromTic, 
                     FLOAT Offset, 
                     SHORT TextRight, 
                     SHORT TextLeft,
                     DOUBLE YScale)
{
   FLOAT y;
   BYTE TestString[PS_MAX_STRING_LENGTH];

   for (y = Start; y <= m_YMax; y *= 10) 
   {
      y = PS_Trunc(y, PS_ROUND_EVEN);
      if ((m_LogTicNumberPosition == PS_AXIS_TIC_AND_NUMBER_TIC_LEFT_BELOW_NUM_LEFT_BELOW) || 
          (m_LogTicNumberPosition == PS_AXIS_TIC_AND_NUMBER_TIC_THROUGH_NUM_LEFT_BELOW) ||
          (m_LogTicNumberPosition == PS_AXIS_TIC_AND_NUMBER_TIC_RIGHT_ABOVE_NUM_LEFT_BELOW))
      {
         fprintf(m_PSFileStream, "%g (%g) ", Offset + TextLeft - NumberDistanceFromTic, y);
         fprintf(m_PSFileStream, "stw pop sub\n");
      }
      else 
      {
         fprintf(m_PSFileStream, "%g ", Offset + TextRight + NumberDistanceFromTic);
      }

      fprintf(m_PSFileStream, "%g ", log10(y / m_YMin) * YScale); 
      fprintf(m_PSFileStream, "charheight 2 div sub mto ");
      fprintf(m_PSFileStream, "(%g) show\n", y);

      sprintf(TestString,"%g", y);
      if (strlen(TestString) > strlen(m_StringAtHighestY))
      {
         sprintf(m_StringAtHighestY, "%g", y);
      }
   }
}

/*********************************************************************************/

void PS_YSetMinNumbers (FLOAT Start, 
                        SHORT NumberDistanceFromTic, 
                        FLOAT Offset, 
                        SHORT TextRight, 
                        SHORT TextLeft,
                        DOUBLE YScale)
{
   FLOAT y;

	y = PS_Trunc(Start, PS_ROUND_EVEN);
   if ((m_LogTicNumberPosition == PS_AXIS_TIC_AND_NUMBER_TIC_LEFT_BELOW_NUM_LEFT_BELOW) || 
       (m_LogTicNumberPosition == PS_AXIS_TIC_AND_NUMBER_TIC_THROUGH_NUM_LEFT_BELOW) ||
       (m_LogTicNumberPosition == PS_AXIS_TIC_AND_NUMBER_TIC_RIGHT_ABOVE_NUM_LEFT_BELOW))
   {
      fprintf(m_PSFileStream, "%g (%g) ", Offset + TextLeft - NumberDistanceFromTic, y);
      fprintf(m_PSFileStream, "stw pop sub\n");
   }
   else 
   {
      fprintf(m_PSFileStream, "%g ", Offset + TextRight + NumberDistanceFromTic);
   }

   fprintf(m_PSFileStream, "%g ", log10(y / m_YMin) * YScale); 
   fprintf(m_PSFileStream, "charheight 2 div sub mto ");
   fprintf(m_PSFileStream, "(%g) show\n", y);
}
