/*****************************************************************/
/*  ELEC 999                                                     */
/*  University of Nebraska-Lincoln                               */
/*  Department of Electrical Engineering                         */
/*  David J. Russell                                             */
/*  5-17-08                                                      */
/*****************************************************************/

#ifdef _PS
#define PS_EXTERN
#else
#define PS_EXTERN extern
#endif

/*
   All code in this module was taken and modified from the Cgraph 
   PostScript plotting library in C [version 2.04, 1998-08-29].

   For the current version, see the original source:
   http://totoro.berkeley.edu/software/A_Cgraph.html

   ftp://pinoko.berkeley.edu/pub/software/Cgraph2.04.tar.gz
   (version number 2.04 may change)

   Izumi Ohzawa, Ph.D.
   izumi@pinoko.berkeley.edu
*/

/******************************************************************/
/* Public Module Constants                                        */
/******************************************************************/

#define PS_POINTS_PER_INCH                            72

/******************************************************************/
/* Public Type Definitions                                        */
/******************************************************************/
typedef enum
{
   PS_INCHES = 0,
   PS_DOMAIN
} PS_COORDINATE_SYSTEM;

typedef enum
{
   PS_LINE_NO_DASH = 0,
   PS_LINE_DASH_1,
   PS_LINE_DASH_2,
   PS_LINE_DASH_3,
   PS_LINE_DASH_4,
   PS_LINE_DASH_5,
   PS_LINE_DASH_6
} PS_LINE_DASH_TYPE;
   
typedef enum
{
   PS_FONT_TIMES = 0,
   PS_FONT_COURIER,
   PS_FONT_HELVETICA
} PS_FONT_TYPE;

typedef enum
{
   PS_FONT_STYLE_NORMAL = 0,
   PS_FONT_STYLE_BOLD,
   PS_FONT_STYLE_ITALIC,
   PS_FONT_STYLE_BOLD_ITALIC
} PS_FONT_STYLE;

typedef enum
{
   PS_MARKER_OPEN_CIRCLE = 0,
   PS_MARKER_FILLED_CIRCLE,
   PS_MARKER_OPEN_TRIANGLE,
   PS_MARKER_FILLED_TRIANGLE,
   PS_MARKER_OPEN_DIAMOND,
   PS_MARKER_FILLED_DIAMOND,
   PS_MARKER_OPEN_SQUARE,
   PS_MARKER_FILLED_SQUARE,
   PS_MARKER_OPEN_STAR,
   PS_MARKER_FILLED_STAR,
   PS_MARKER_OPEN_PENTAGON,
   PS_MARKER_FILLED_PENTAGON,
   PS_MARKER_OPEN_INVERTED_TRIANGLE,
   PS_MARKER_FILLED_INVERTED_TRIANGLE,
   PS_MARKER_BOWTIE,
   PS_MARKER_JIGGER,
   PS_MARKER_ASTERISK,
   PS_MARKER_CROSSHAIRS,
   PS_MARKER_SPINNING_WHEEL,
   PS_MARKER_CROSSHATCH,
   PS_MARKER_MAX_NUMBER                /* Not a marker, just marking the end. */
} PS_MARKER_TYPE;

typedef enum
{
   PS_POINTER_RIGHT = 0,
   PS_POINTER_DOWN,
   PS_POINTER_LEFT,
   PS_POINTER_UP
} PS_POINTER_TYPE;

typedef enum
{
   PS_AXIS_TIC_AND_NUMBER_TIC_LEFT_BELOW_NUM_LEFT_BELOW = 1,
   PS_AXIS_TIC_AND_NUMBER_TIC_THROUGH_NUM_LEFT_BELOW,
   PS_AXIS_TIC_AND_NUMBER_TIC_RIGHT_ABOVE_NUM_LEFT_BELOW,
   PS_AXIS_TIC_AND_NUMBER_TIC_RIGHT_ABOVE_NUM_RIGHT_ABOVE,
   PS_AXIS_TIC_AND_NUMBER_TIC_THROUGH_NUM_RIGHT_ABOVE,
   PS_AXIS_TIC_AND_NUMBER_TIC_LEFT_BELOW_NUM_RIGHT_ABOVE
} PS_AXIS_TIC_AND_NUMBER_TYPE;

/*********************************************************************************/
/* Public Module Variables */
/*********************************************************************************/

/*********************************************************************************/
/* Public Module Function Prototypes */
/*********************************************************************************/

void PS_SetDefaults (FLOAT Scale);
void PS_Initialize (BOOLEAN UseLandscape, BYTE *TopCommentString, USHORT Width, USHORT Height);
void PS_ShowPage (void);
void PS_SetDashPattern (PS_LINE_DASH_TYPE Type, FLOAT DashPeriod);
void PS_CenterLabel (BYTE *TextString, FLOAT X, FLOAT Y , FLOAT RotateBy, PS_COORDINATE_SYSTEM CoordinateSystem);
void PS_XLabel (BYTE *TextString);
void PS_YLabel (BYTE *TextString);
void PS_TextAlign (FLOAT X, FLOAT Y, FLOAT RotateBy, BYTE *TextString, SHORT XPos, SHORT YPos, PS_COORDINATE_SYSTEM CoordinateSystem);
void PS_SelectCoordinate (PS_COORDINATE_SYSTEM Marks, PS_COORDINATE_SYSTEM Text);
void PS_Font (FLOAT Size, PS_FONT_TYPE Type, PS_FONT_STYLE Style);
void PS_FontByName (FLOAT Size, BYTE *Name);
void PS_SetGray (FLOAT Degree);
void PS_SetRGBColor (FLOAT Red, FLOAT Green, FLOAT Blue);
void PS_SetGrayRGBColor(FLOAT Degree, FLOAT Red, FLOAT Green, FLOAT Blue);
void PS_SetMarkerGray (FLOAT NewPerimeter, FLOAT NewInside);
void PS_SetMarkerColor (FLOAT NewPerimeterRed, 
                        FLOAT NewPerimeterGreen, 
                        FLOAT NewPerimeterBlue, 
                        FLOAT NewInsideRed, 
                        FLOAT NewInsideGreen, 
                        FLOAT NewInsideBlue);
void PS_SetMarkerGrayRGBColor (FLOAT NewPerimeterGray, 
                        FLOAT NewPerimeterRed, 
                        FLOAT NewPerimeterGreen, 
                        FLOAT NewPerimeterBlue, 
                        FLOAT NewInsideGray, 
                        FLOAT NewInsideRed, 
                        FLOAT NewInsideGreen, 
                        FLOAT NewInsideBlue);
void PS_UseColor (BOOLEAN Flag);
void PS_UseOpaqueBG (BOOLEAN Flag, FLOAT Gray, FLOAT Red, FLOAT Green, FLOAT Blue);
void PS_Marker (FLOAT X, FLOAT Y, PS_MARKER_TYPE Type, FLOAT Size);
void PS_Pointer (FLOAT X, FLOAT Y, PS_POINTER_TYPE Type, FLOAT Size);
void PS_Mesh (void);
void PS_Move (FLOAT XPos, FLOAT YPos);
void PS_Line (FLOAT XPos, FLOAT YPos);
void PS_RectFill (FLOAT XPos, FLOAT YPos, FLOAT Width, FLOAT Height);
void PS_ShiftOriginRelative (FLOAT XPos, FLOAT YPos);
void PS_ShiftOriginAbsolute (FLOAT XPos, FLOAT YPos);
void PS_Fill (void);
void PS_EveryOtherFill (void);
void PS_ClosePath (void);
void PS_GraphicsStateSave (void);
void PS_GraphicsStateRestore (void);
void PS_InsertPostScriptCommand (BYTE *PostScriptCommand);
void PS_MoveRelative (FLOAT XPos, FLOAT YPos);
void PS_LineRelative (FLOAT XPos, FLOAT YPos);
void PS_Reset (void);
void PS_Stroke (void);
void PS_AxisEnable (BOOLEAN AxisEnable, BOOLEAN AxisNumberEnable);
void PS_LogAxisStyle (USHORT NumberSelectMask, 
                        USHORT TicSelectMask,
                        UBYTE NumberDistanceFromTic,
                        PS_AXIS_TIC_AND_NUMBER_TYPE TicNumberPosition,
                        UBYTE TicLength,
                        UBYTE TicLength10);
void PS_LinearAxisStyle (UBYTE Precision, 
                           UBYTE TicOffsetNumbering, 
                           UBYTE NumberDistanceFromTic, 
                           PS_AXIS_TIC_AND_NUMBER_TYPE TicNumberPosition, 
                           UBYTE TicLength);
void PS_Text (FLOAT XPos, FLOAT YPos, FLOAT Direction, BYTE *Text);
void PS_LineWidth (FLOAT LineWidth);
void PS_XAxis (FLOAT XSize, FLOAT XMin, FLOAT XMax, FLOAT Offset, FLOAT XTicSeparation, UBYTE NumberSeparation);
void PS_YAxis (FLOAT YSize, FLOAT YMin, FLOAT YMax, FLOAT Offset, FLOAT YTicSeparation, UBYTE NumberSeparation);
void PS_XLog (FLOAT XSize, FLOAT XMin, FLOAT XMax, FLOAT Offset);
void PS_YLog (FLOAT YSize, FLOAT YMin, FLOAT YMax, FLOAT Offset);

