/*****************************************************************/
/*  ELEC 999                                                     */
/*  University of Nebraska-Lincoln                               */
/*  Department of Electrical Engineering                         */
/*  David J. Russell                                             */
/*  5-24-08                                                      */
/*                                                               */
/* 02/16/09 - Added minimum ID parameter for labeling nodes on   */
/*             .eps outputs, version 1.6                         */
/*****************************************************************/

#define _PST

/*********************************************************************************/
/* Included Header Files */
/*********************************************************************************/
#include "Globals.h"
#include "PostScript.h"
#include "StringADT.h"
#include "PostScriptTree.h"

/*********************************************************************************/
/* Private Module Constants */
/*********************************************************************************/

#define PST_DEFAULT_PAGE_SCALE                  1.0
#define PST_DEFAULT_XAXIS_MAX_INCHES            10.0
#define PST_DEFAULT_YAXIS_MAX_INCHES            7.5
#define PST_DEFAULT_ORIGIN_PAGE_BOUNDARY_INCHES 0.50
#define PST_DEFAULT_XAXIS_MAX_POINTS            (PS_POINTS_PER_INCH * PST_DEFAULT_XAXIS_MAX_INCHES)
#define PST_DEFAULT_YAXIS_MAX_POINTS            (PS_POINTS_PER_INCH * PST_DEFAULT_YAXIS_MAX_INCHES)
#define PST_DEFAULT_PAGE_WIDTH_POINTS           (PS_POINTS_PER_INCH * (PST_DEFAULT_XAXIS_MAX_INCHES + (PST_DEFAULT_ORIGIN_PAGE_BOUNDARY_INCHES * 2)))
#define PST_DEFAULT_PAGE_HEIGHT_POINTS          (PS_POINTS_PER_INCH * (PST_DEFAULT_YAXIS_MAX_INCHES + (PST_DEFAULT_ORIGIN_PAGE_BOUNDARY_INCHES * 2)))
#define PST_DEFAULT_LINE_COLOR                  PST_COLOR_BLACK
#define PST_DEFAULT_LINE_WIDTH                  1.0
#define PST_DEFAULT_MARKER_SIZE                 10.0
#define PST_DEFAULT_NO_LABEL_MARKER_SIZE        5.0

#define PST_DEFAULT_MARKERS_PER_INCH            (PS_POINTS_PER_INCH / (PST_DEFAULT_MARKER_SIZE * 2))

#define PST_MARKER_TEXT_OFFSET_X                ((PST_DEFAULT_MARKER_SIZE / 2) + 2)
#define PST_MARKER_TEXT_OFFSET_Y                ((PST_DEFAULT_MARKER_SIZE / 2) - 2)

#define PST_ASTERISK                            0x2A
#define PST_APOSTROPHE                          0x27
#define PST_OPEN_PARENTHESIS                    0x28
#define PST_CLOSE_PARENTHESIS                   0x29
#define PST_QUOTATION_MARK                      0x22
#define PST_COLON                               0x3A
#define PST_ZERO                                0x30
#define PST_NINE                                0x39
#define PST_COMMA                               0x2C

#define PST_NO_LABEL                            0xFFFFFFFF
#define PST_MAX_ULONG_STRING                    12

/*********************************************************************************/
/* Private Type Definitions */
/*********************************************************************************/
typedef enum
{
   PST_NODE_NORMAL,
   PST_NODE_REVERSE_COMPLEMENT,
   PST_NODE_SELF_COMPLEMENT,
   PST_NODE_ROOT
} PST_NODE_TYPE;

typedef enum
{
   PST_COLOR_WHITE,
   PST_COLOR_BLACK,
   PST_COLOR_BLUE,
   PST_COLOR_GREEN,
   PST_COLOR_RED,
   PST_COLOR_YELLOW,
   PST_COLOR_MAGENTA,
   PST_COLOR_CYAN,
   PST_NUMBER_OF_COLORS
} PST_COLOR_TYPE;

typedef struct
{
   FLOAT Red;
   FLOAT Green;
   FLOAT Blue;
} PST_COLOR_COMPONENT_TYPE;

typedef void (*PST_POINTER_TO_STATE_FUNCTION) (UBYTE, UBYTE *);

typedef enum
{
   PST_STATE_ROOT,
   PST_STATE_LINE_DOWN_LEFT,
   PST_STATE_MOVE_UP_RIGHT,
   PST_STATE_LINE_DOWN_RIGHT,
   PST_STATE_MOVE_UP_LEFT,
} PST_STATE;

typedef struct
{
   PST_STATE State;
   PST_POINTER_TO_STATE_FUNCTION StateFunction;
} PST_STATE_ENTRY;

/*********************************************************************************/
/* Private Module Function Prototypes */
/*********************************************************************************/
static void PST_Initialize (ULONG FrontierLength);
static void PST_MoveToRootPosition (void);
static void PST_LineDownLeft (ULONG Depths);
static void PST_LineDownRight (ULONG Depths);
static void PST_MarkThenMoveUpLeft (ULONG Depths, ULONG Label, PST_NODE_TYPE NodeType);
static void PST_MarkThenMoveUpRight (ULONG Depths, ULONG Label, PST_NODE_TYPE NodeType);
static void PST_MarkCurrentPositionAsRoot (void);
static void PST_PlaceMark (ULONG Label, PST_NODE_TYPE NodeType);
static void PST_PlaceMarkText (ULONG Label, PST_NODE_TYPE NodeType);

static void PST_InitializeULongStack (ULONG FrontierLength, ULONG **Stack, ULONG *StackIndex);
static void PST_CleanUpULongStack (ULONG **Stack, ULONG *StackIndex);
static void PST_PushULongStack (ULONG Value, ULONG *Stack, ULONG *StackIndex);
static ULONG PST_PopULongStack (ULONG *Stack, ULONG *StackIndex);

static void PST_InitializeNodeTypeStack (ULONG FrontierLength, PST_NODE_TYPE **Stack, ULONG *StackIndex);
static void PST_CleanUpNodeTypeStack (PST_NODE_TYPE **Stack, ULONG *StackIndex);
static void PST_PushNodeTypeStack (PST_NODE_TYPE Value, PST_NODE_TYPE *Stack, ULONG *StackIndex);
static PST_NODE_TYPE PST_PopNodeTypeStack (PST_NODE_TYPE *Stack, ULONG *StackIndex);

static void PST_FindStartOfTreeString(UBYTE *TreeString);
static void PST_ProcessTreeString (UBYTE *TreeString);
static ULONG PST_FindIndexOfMatchingToken (UBYTE Token, 
                                          UBYTE *TreeString, 
                                          ULONG CurrentTreeStringIndex);
                                          
static void PST_FindDepthAndLabelAfterMatchingParenthesis (UBYTE *TreeString, 
                                                            ULONG *Depth, 
                                                            ULONG *Label, 
                                                            PST_NODE_TYPE *NodeType);
static BOOLEAN PST_FindNextDepthAndLabel (UBYTE *TreeString, 
                                 ULONG CurrentTreeStringIndex, 
                                 BOOLEAN MoveGlobalIndex, 
                                 ULONG *Depth, 
                                 ULONG *Label, 
                                 PST_NODE_TYPE *NodeType);

/* States */
static void PST_StateRoot (UBYTE Ch, UBYTE *TreeString);
static void PST_StateLineDownLeft(UBYTE Ch, UBYTE *TreeString);
static void PST_StateMoveUpRight(UBYTE Ch, UBYTE *TreeString);
static void PST_StateLineDownRight(UBYTE Ch, UBYTE *TreeString);
static void PST_StateMoveUpLeft(UBYTE Ch, UBYTE *TreeString);

/*********************************************************************************/
/* Private Module Variables */
/*********************************************************************************/
static ULONG m_TreeDepth;
static FLOAT m_VerticalStepSize;
static FLOAT m_HorizontalStepSize;
static FLOAT m_RootXPosition;
static FLOAT m_RootYPosition;
static FLOAT m_CurrentXPosition;
static FLOAT m_CurrentYPosition;

static ULONG m_CurrentTreeStringIndex;
static ULONG m_CurrentParenthesisCount;

static ULONG *m_DepthStack;
static ULONG m_DepthStackPointer;

static ULONG *m_LabelStack;
static ULONG m_LabelStackPointer;

static PST_NODE_TYPE *m_NodeTypeStack;
static ULONG m_NodeTypeStackPointer;

const static PST_COLOR_COMPONENT_TYPE m_Colors[PST_NUMBER_OF_COLORS] =
{
   {1.0, 1.0, 1.0},
   {0.0, 0.0, 0.0},
   {0.0, 0.0, 1.0},
   {0.0, 1.0, 0.0},
   {1.0, 0.0, 0.0},
   {1.0, 1.0, 0.0},
   {1.0, 0.0, 1.0},
   {0.0, 1.0, 1.0},
};

static PST_STATE m_State;

const static PST_STATE_ENTRY m_StateTable[] =
{
   {PST_STATE_ROOT, PST_StateRoot},
   {PST_STATE_LINE_DOWN_LEFT, PST_StateLineDownLeft},
   {PST_STATE_MOVE_UP_RIGHT, PST_StateMoveUpRight},
   {PST_STATE_LINE_DOWN_RIGHT, PST_StateLineDownRight},
   {PST_STATE_MOVE_UP_LEFT, PST_StateMoveUpLeft},
};

/*********************************************************************************/
/* Public and Private Module Functions */
/*********************************************************************************/

void PST_ParseTreeStringIntoPostScript (STRINGADT_STRING_TYPE *TreeString, ULONG MaxDepth)
{
   PST_Initialize (MaxDepth);
   PST_InitializeULongStack (MaxDepth, &m_DepthStack, &m_DepthStackPointer);
   PST_InitializeULongStack (MaxDepth, &m_LabelStack, &m_LabelStackPointer);
   PST_InitializeNodeTypeStack (MaxDepth, &m_NodeTypeStack, &m_NodeTypeStackPointer);

   PST_MoveToRootPosition();

   PST_FindStartOfTreeString(TreeString->Characters);
   PST_ProcessTreeString(TreeString->Characters);
   
   PS_ShowPage();

   PST_CleanUpNodeTypeStack (&m_NodeTypeStack, &m_NodeTypeStackPointer);
   PST_CleanUpULongStack (&m_LabelStack, &m_LabelStackPointer);
   PST_CleanUpULongStack (&m_DepthStack, &m_DepthStackPointer);
}

/*********************************************************************************/

void PST_Initialize (ULONG FrontierLength)
{
   DOUBLE WidthPoints;
   DOUBLE HeightPoints;
   DOUBLE XAxisMaxInches;
   DOUBLE YAxisMaxInches;
   DOUBLE XAxisMaxPoints;
   DOUBLE YAxisMaxPoints;

   WidthPoints = (((DOUBLE) FrontierLength) / PST_DEFAULT_MARKERS_PER_INCH) * PS_POINTS_PER_INCH;
   if (WidthPoints < PST_DEFAULT_PAGE_WIDTH_POINTS)
   {
      WidthPoints = PST_DEFAULT_PAGE_WIDTH_POINTS;
   }
   HeightPoints = (WidthPoints / PST_DEFAULT_PAGE_WIDTH_POINTS) * PST_DEFAULT_PAGE_HEIGHT_POINTS;

   XAxisMaxInches = (WidthPoints / PS_POINTS_PER_INCH) - (PST_DEFAULT_ORIGIN_PAGE_BOUNDARY_INCHES * 2);
   XAxisMaxPoints = XAxisMaxInches * PS_POINTS_PER_INCH;
   YAxisMaxInches = (HeightPoints / PS_POINTS_PER_INCH) - (PST_DEFAULT_ORIGIN_PAGE_BOUNDARY_INCHES * 2);
   YAxisMaxPoints = YAxisMaxInches * PS_POINTS_PER_INCH;
   
   PS_SetDefaults (PST_DEFAULT_PAGE_SCALE);
   PS_UseColor (TRUE);
   PS_Initialize (FALSE, NULL, WidthPoints, HeightPoints);

   PS_SelectCoordinate(PS_DOMAIN, PS_DOMAIN);

   PS_SetRGBColor(m_Colors[PST_DEFAULT_LINE_COLOR].Red,
                  m_Colors[PST_DEFAULT_LINE_COLOR].Green,
                  m_Colors[PST_DEFAULT_LINE_COLOR].Blue);

   PS_ShiftOriginAbsolute (PST_DEFAULT_ORIGIN_PAGE_BOUNDARY_INCHES, 
                           PST_DEFAULT_ORIGIN_PAGE_BOUNDARY_INCHES);
   PS_AxisEnable (FALSE, FALSE);
   PS_LinearAxisStyle (0, 0, 0, 1, 0);
   PS_LineWidth (PST_DEFAULT_LINE_WIDTH);
   
   PS_XAxis (XAxisMaxInches, 0, XAxisMaxPoints, 0, 0, 0);
   PS_YAxis (YAxisMaxInches, 0, YAxisMaxPoints, 0, 0, 0);
   
   m_TreeDepth = FrontierLength;

   m_RootXPosition = XAxisMaxPoints / 2.0;
   m_RootYPosition = YAxisMaxPoints;

   if (m_TreeDepth > 1)
   {
      m_VerticalStepSize = YAxisMaxPoints / (m_TreeDepth - 1);
      m_HorizontalStepSize = (XAxisMaxPoints / (m_TreeDepth - 1)) / 2;
   }
}

/*********************************************************************************/

void PST_MoveToRootPosition (void)
{
   m_CurrentXPosition = m_RootXPosition;
   m_CurrentYPosition = m_RootYPosition;
   PS_Move (m_CurrentXPosition, m_CurrentYPosition);   
}

/*********************************************************************************/

void PST_LineDownLeft (ULONG Depths)
{
   m_CurrentXPosition -= m_HorizontalStepSize * Depths;
   m_CurrentYPosition -= m_VerticalStepSize * Depths;
   PS_Line (m_CurrentXPosition, m_CurrentYPosition);
   PS_Stroke();
   PS_Move (m_CurrentXPosition, m_CurrentYPosition);   
}

/*********************************************************************************/

void PST_LineDownRight (ULONG Depths)
{
   m_CurrentXPosition += m_HorizontalStepSize * Depths;
   m_CurrentYPosition -= m_VerticalStepSize * Depths;
   PS_Line (m_CurrentXPosition, m_CurrentYPosition);
   PS_Stroke();
   PS_Move (m_CurrentXPosition, m_CurrentYPosition);   
}

/*********************************************************************************/

void PST_MarkThenMoveUpLeft (ULONG Depths, ULONG Label, PST_NODE_TYPE NodeType)
{
   PST_PlaceMark (Label, NodeType);
   PST_PlaceMarkText (Label, NodeType);
   m_CurrentXPosition -= m_HorizontalStepSize * Depths;
   m_CurrentYPosition += m_VerticalStepSize * Depths;
   PS_Move (m_CurrentXPosition, m_CurrentYPosition);
}

/*********************************************************************************/

void PST_MarkThenMoveUpRight (ULONG Depths, ULONG Label, PST_NODE_TYPE NodeType)
{
   PST_PlaceMark (Label, NodeType);
   PST_PlaceMarkText (Label, NodeType);
   m_CurrentXPosition += m_HorizontalStepSize * Depths;
   m_CurrentYPosition += m_VerticalStepSize * Depths;
   PS_Move (m_CurrentXPosition, m_CurrentYPosition);
}

/*********************************************************************************/

void PST_MarkCurrentPositionAsRoot (void)
{
   PST_PlaceMark (0, PST_NODE_ROOT);
   PST_PlaceMarkText (0, PST_NODE_ROOT);
}

/*********************************************************************************/

void PST_PlaceMark (ULONG Label, PST_NODE_TYPE NodeType)
{
   if (Label != PST_NO_LABEL)
   {
      switch (NodeType)
      {
         case PST_NODE_NORMAL:
         case PST_NODE_REVERSE_COMPLEMENT:
         case PST_NODE_SELF_COMPLEMENT:
            PS_SetMarkerColor (m_Colors[PST_COLOR_BLACK].Red,
                                 m_Colors[PST_COLOR_BLACK].Green,
                                 m_Colors[PST_COLOR_BLACK].Blue, 
                                 m_Colors[Label % PST_NUMBER_OF_COLORS].Red,
                                 m_Colors[Label % PST_NUMBER_OF_COLORS].Green,
                                 m_Colors[Label % PST_NUMBER_OF_COLORS].Blue);
            break;
         
         case PST_NODE_ROOT:
            PS_SetMarkerColor (m_Colors[PST_COLOR_BLACK].Red,
                                 m_Colors[PST_COLOR_BLACK].Green,
                                 m_Colors[PST_COLOR_BLACK].Blue, 
                                 m_Colors[PST_COLOR_WHITE].Red,
                                 m_Colors[PST_COLOR_WHITE].Green,
                                 m_Colors[PST_COLOR_WHITE].Blue);
            break;
      }

      switch (NodeType)
      {
         case PST_NODE_NORMAL:
            PS_Marker (m_CurrentXPosition, m_CurrentYPosition, PS_MARKER_OPEN_TRIANGLE, PST_DEFAULT_MARKER_SIZE);
            break;
            
         case PST_NODE_REVERSE_COMPLEMENT:
            PS_Marker (m_CurrentXPosition, m_CurrentYPosition, PS_MARKER_OPEN_INVERTED_TRIANGLE, PST_DEFAULT_MARKER_SIZE);
            break;
            
         case PST_NODE_SELF_COMPLEMENT:
            PS_Marker (m_CurrentXPosition, m_CurrentYPosition, PS_MARKER_OPEN_PENTAGON, PST_DEFAULT_MARKER_SIZE);
            break;
         
         case PST_NODE_ROOT:
            PS_Marker (m_CurrentXPosition, m_CurrentYPosition, PS_MARKER_OPEN_CIRCLE, PST_DEFAULT_MARKER_SIZE);
            break;
      }
   }
   else
   {
      PS_SetMarkerColor (m_Colors[PST_COLOR_BLACK].Red,
                           m_Colors[PST_COLOR_BLACK].Green,
                           m_Colors[PST_COLOR_BLACK].Blue, 
                           m_Colors[PST_COLOR_WHITE].Red,
                           m_Colors[PST_COLOR_WHITE].Green,
                           m_Colors[PST_COLOR_WHITE].Blue);
      PS_Marker (m_CurrentXPosition, m_CurrentYPosition, PS_MARKER_OPEN_SQUARE, PST_DEFAULT_NO_LABEL_MARKER_SIZE);
   }
}

/*********************************************************************************/

void PST_PlaceMarkText (ULONG Label, PST_NODE_TYPE NodeType)
{
   BYTE LabelString[PST_MAX_ULONG_STRING];
   
   if (Label != PST_NO_LABEL)
   {
      switch (NodeType)
      {
         case PST_NODE_NORMAL:
            sprintf(LabelString, "%lu", Label);
            break;
            
         case PST_NODE_REVERSE_COMPLEMENT:
            sprintf(LabelString, "%lu'", Label);
            break;
            
         case PST_NODE_SELF_COMPLEMENT:
            sprintf(LabelString, "%lu*", Label);
            break;

         case PST_NODE_ROOT:
            sprintf(LabelString, "Start");
            break;
      }

      PS_Text (m_CurrentXPosition + PST_MARKER_TEXT_OFFSET_X, m_CurrentYPosition - PST_MARKER_TEXT_OFFSET_Y, 0, LabelString);
   }
}

/*********************************************************************************/

void PST_InitializeULongStack (ULONG FrontierLength, ULONG **Stack, ULONG *StackIndex)
{
   *Stack = (ULONG *) malloc(FrontierLength * sizeof(ULONG));

   if (*Stack == NULL)
   {
      printf("Error: malloc failed in PST_InitializeULongStack.\n");
      exit(1);
   }
   
   *StackIndex = 0;
}

/*********************************************************************************/

void PST_CleanUpULongStack (ULONG **Stack, ULONG *StackIndex)
{
   if (*Stack != NULL)
   {
      free(*Stack);
      *Stack = NULL;
   }
}

/*********************************************************************************/

void PST_PushULongStack (ULONG Value, ULONG *Stack, ULONG *StackIndex)
{
   Stack[*StackIndex] = Value;
   *StackIndex = *StackIndex + 1;
}

/*********************************************************************************/

ULONG PST_PopULongStack (ULONG *Stack, ULONG *StackIndex)
{
   ULONG Value;
   
   if (*StackIndex > 0)
   {
      *StackIndex = *StackIndex - 1;
      Value = Stack[*StackIndex];
   }
   else
   {
      printf("Error: popping empty ULong stack.\n");
      exit(1);
   }
   
   return (Value);
}

/*********************************************************************************/

void PST_InitializeNodeTypeStack (ULONG FrontierLength, PST_NODE_TYPE **Stack, ULONG *StackIndex)
{
   *Stack = (PST_NODE_TYPE *) malloc(FrontierLength * sizeof(PST_NODE_TYPE));

   if (*Stack == NULL)
   {
      printf("Error: malloc failed in PST_InitializeNodeTypeStack.\n");
      exit(1);
   }
   
   *StackIndex = 0;
}

/*********************************************************************************/

void PST_CleanUpNodeTypeStack (PST_NODE_TYPE **Stack, ULONG *StackIndex)
{
   if (*Stack != NULL)
   {
      free(*Stack);
      *Stack = NULL;
   }
}

/*********************************************************************************/

void PST_PushNodeTypeStack (PST_NODE_TYPE Value, PST_NODE_TYPE *Stack, ULONG *StackIndex)
{
   Stack[*StackIndex] = Value;
   *StackIndex = *StackIndex + 1;
}

/*********************************************************************************/

PST_NODE_TYPE PST_PopNodeTypeStack (PST_NODE_TYPE *Stack, ULONG *StackIndex)
{
   PST_NODE_TYPE Value;
   
   if (*StackIndex > 0)
   {
      *StackIndex = *StackIndex - 1;
      Value = Stack[*StackIndex];
   }
   else
   {
      printf("Error: popping empty Boolean stack.\n");
      exit(1);
   }
   
   return (Value);
}

/*********************************************************************************/

void PST_FindStartOfTreeString(UBYTE *TreeString)
{
   m_CurrentTreeStringIndex = 0;
   m_CurrentParenthesisCount = 0;

   if (TreeString[m_CurrentTreeStringIndex] != PST_OPEN_PARENTHESIS)
   {
      printf("Error: tree string is invalid.\n");
      exit(1);
   }
   
   m_CurrentTreeStringIndex++;
   m_CurrentParenthesisCount++;
   m_State = PST_STATE_ROOT;
}

/*********************************************************************************/

void PST_ProcessTreeString (UBYTE *TreeString)
{
   UBYTE Ch;
   
   Ch = TreeString[m_CurrentTreeStringIndex];
   while ((m_CurrentParenthesisCount > 0) && (Ch != 0))
   {
      Ch = TreeString[m_CurrentTreeStringIndex++];
      m_StateTable[m_State].StateFunction(Ch, TreeString);
   }
   
   if (Ch == 0)
   {
      printf("Error: parenthesis do not match in tree string.\n");
      exit(1);
   }
   else
   {
      PST_MarkCurrentPositionAsRoot();
   }
}

/*********************************************************************************/

void PST_StateRoot (UBYTE Ch, UBYTE *TreeString)
{
   ULONG Depth;
   ULONG Label;
   PST_NODE_TYPE NodeType;
   
   if (!isspace(Ch))
   {
      switch(Ch)
      {
         case PST_OPEN_PARENTHESIS:
            PST_FindDepthAndLabelAfterMatchingParenthesis (TreeString, &Depth, &Label, &NodeType);
            PST_LineDownLeft(Depth);
            PST_PushULongStack(Depth, m_DepthStack, &m_DepthStackPointer);
            PST_PushULongStack(Label, m_LabelStack, &m_LabelStackPointer);
            PST_PushNodeTypeStack(NodeType, m_NodeTypeStack, &m_NodeTypeStackPointer);
            m_CurrentParenthesisCount++;
            m_State = PST_STATE_LINE_DOWN_LEFT;            
            break;
            
         case PST_QUOTATION_MARK:
            PST_FindNextDepthAndLabel (TreeString, m_CurrentTreeStringIndex - 1, FALSE, &Depth, &Label, &NodeType);
            PST_LineDownLeft(Depth);
            PST_PushULongStack(Depth, m_DepthStack, &m_DepthStackPointer);
            PST_PushULongStack(Label, m_LabelStack, &m_LabelStackPointer);
            PST_PushNodeTypeStack(NodeType, m_NodeTypeStack, &m_NodeTypeStackPointer);
            m_State = PST_STATE_LINE_DOWN_LEFT;            
            break;
            
         case PST_COLON:
            PST_FindNextDepthAndLabel (TreeString, m_CurrentTreeStringIndex - 1, TRUE, &Depth, &Label, &NodeType);
            PST_LineDownLeft(Depth);
            PST_PushULongStack(Depth, m_DepthStack, &m_DepthStackPointer);
            PST_PushULongStack(Label, m_LabelStack, &m_LabelStackPointer);
            PST_PushNodeTypeStack(NodeType, m_NodeTypeStack, &m_NodeTypeStackPointer);
            m_State = PST_STATE_LINE_DOWN_LEFT;            
            break;
         
         default:
            printf("Error: tree string is invalid (1:%c).\n", Ch);
            exit(1);
            break;
      }
   }
}

/*********************************************************************************/

void PST_StateLineDownLeft (UBYTE Ch, UBYTE *TreeString)
{
   ULONG Depth;
   ULONG Label;
   PST_NODE_TYPE NodeType;
   
   if (!isspace(Ch))
   {
      switch(Ch)
      {
         case PST_OPEN_PARENTHESIS:
            PST_FindDepthAndLabelAfterMatchingParenthesis (TreeString, &Depth, &Label, &NodeType);
            PST_LineDownLeft(Depth);
            PST_PushULongStack(Depth, m_DepthStack, &m_DepthStackPointer);
            PST_PushULongStack(Label, m_LabelStack, &m_LabelStackPointer);
            PST_PushNodeTypeStack(NodeType, m_NodeTypeStack, &m_NodeTypeStackPointer);
            m_CurrentParenthesisCount++;
            m_State = PST_STATE_LINE_DOWN_LEFT;            
            break;
            
         case PST_QUOTATION_MARK:
            PST_FindNextDepthAndLabel (TreeString, m_CurrentTreeStringIndex - 1, TRUE, &Depth, &Label, &NodeType);
            PST_LineDownLeft(Depth);
            PST_PushULongStack(Depth, m_DepthStack, &m_DepthStackPointer);
            PST_PushULongStack(Label, m_LabelStack, &m_LabelStackPointer);
            PST_PushNodeTypeStack(NodeType, m_NodeTypeStack, &m_NodeTypeStackPointer);
            m_State = PST_STATE_LINE_DOWN_LEFT;            
            break;
            
         case PST_COLON:
            PST_FindNextDepthAndLabel (TreeString, m_CurrentTreeStringIndex - 1, TRUE, &Depth, &Label, &NodeType);
            PST_LineDownLeft(Depth);
            PST_PushULongStack(Depth, m_DepthStack, &m_DepthStackPointer);
            PST_PushULongStack(Label, m_LabelStack, &m_LabelStackPointer);
            PST_PushNodeTypeStack(NodeType, m_NodeTypeStack, &m_NodeTypeStackPointer);
            m_State = PST_STATE_LINE_DOWN_LEFT;            
            break;
         
         case PST_COMMA:
            Depth = PST_PopULongStack(m_DepthStack, &m_DepthStackPointer);
            Label = PST_PopULongStack(m_LabelStack, &m_LabelStackPointer);
            NodeType = PST_PopNodeTypeStack(m_NodeTypeStack, &m_NodeTypeStackPointer);
            PST_MarkThenMoveUpRight(Depth, Label, NodeType);
            m_State = PST_STATE_MOVE_UP_RIGHT;            
            break;
         
         default:
            printf("Error: tree string is invalid (2:%c).\n", Ch);
            exit(1);
            break;
      }
   }
}

/*********************************************************************************/

void PST_StateMoveUpRight (UBYTE Ch, UBYTE *TreeString)
{
   ULONG Depth;
   ULONG Label;
   PST_NODE_TYPE NodeType;
   
   if (!isspace(Ch))
   {
      switch(Ch)
      {
         case PST_OPEN_PARENTHESIS:
            PST_FindDepthAndLabelAfterMatchingParenthesis (TreeString, &Depth, &Label, &NodeType);
            PST_LineDownRight(Depth);
            PST_PushULongStack(Depth, m_DepthStack, &m_DepthStackPointer);
            PST_PushULongStack(Label, m_LabelStack, &m_LabelStackPointer);
            PST_PushNodeTypeStack(NodeType, m_NodeTypeStack, &m_NodeTypeStackPointer);
            m_CurrentParenthesisCount++;
            m_State = PST_STATE_LINE_DOWN_RIGHT;            
            break;
            
         case PST_QUOTATION_MARK:
            PST_FindNextDepthAndLabel (TreeString, m_CurrentTreeStringIndex - 1, TRUE, &Depth, &Label, &NodeType);
            PST_LineDownRight(Depth);
            PST_PushULongStack(Depth, m_DepthStack, &m_DepthStackPointer);
            PST_PushULongStack(Label, m_LabelStack, &m_LabelStackPointer);
            PST_PushNodeTypeStack(NodeType, m_NodeTypeStack, &m_NodeTypeStackPointer);
            m_State = PST_STATE_LINE_DOWN_RIGHT;            
            break;
            
         case PST_COLON:
            PST_FindNextDepthAndLabel (TreeString, m_CurrentTreeStringIndex - 1, TRUE, &Depth, &Label, &NodeType);
            PST_LineDownRight(Depth);
            PST_PushULongStack(Depth, m_DepthStack, &m_DepthStackPointer);
            PST_PushULongStack(Label, m_LabelStack, &m_LabelStackPointer);
            PST_PushNodeTypeStack(NodeType, m_NodeTypeStack, &m_NodeTypeStackPointer);
            m_State = PST_STATE_LINE_DOWN_RIGHT;            
            break;
         
         default:
            printf("Error: tree string is invalid (3:%c).\n", Ch);
            exit(1);
            break;
      }
   }
}

/*********************************************************************************/

void PST_StateLineDownRight (UBYTE Ch, UBYTE *TreeString)
{
   ULONG Depth;
   ULONG Label;
   PST_NODE_TYPE NodeType;
   
   if (!isspace(Ch))
   {
      switch(Ch)
      {
         case PST_OPEN_PARENTHESIS:
            PST_FindDepthAndLabelAfterMatchingParenthesis (TreeString, &Depth, &Label, &NodeType);
            PST_LineDownLeft(Depth);
            PST_PushULongStack(Depth, m_DepthStack, &m_DepthStackPointer);
            PST_PushULongStack(Label, m_LabelStack, &m_LabelStackPointer);
            PST_PushNodeTypeStack(NodeType, m_NodeTypeStack, &m_NodeTypeStackPointer);
            m_CurrentParenthesisCount++;
            m_State = PST_STATE_LINE_DOWN_LEFT;            
            break;
            
         case PST_QUOTATION_MARK:
            PST_FindNextDepthAndLabel (TreeString, m_CurrentTreeStringIndex - 1, TRUE, &Depth, &Label, &NodeType);
            PST_LineDownLeft(Depth);
            PST_PushULongStack(Depth, m_DepthStack, &m_DepthStackPointer);
            PST_PushULongStack(Label, m_LabelStack, &m_LabelStackPointer);
            PST_PushNodeTypeStack(NodeType, m_NodeTypeStack, &m_NodeTypeStackPointer);
            m_State = PST_STATE_LINE_DOWN_LEFT;            
            break;
            
         case PST_COLON:
            PST_FindNextDepthAndLabel (TreeString, m_CurrentTreeStringIndex - 1, TRUE, &Depth, &Label, &NodeType);
            PST_LineDownLeft(Depth);
            PST_PushULongStack(Depth, m_DepthStack, &m_DepthStackPointer);
            PST_PushULongStack(Label, m_LabelStack, &m_LabelStackPointer);
            PST_PushNodeTypeStack(NodeType, m_NodeTypeStack, &m_NodeTypeStackPointer);
            m_State = PST_STATE_LINE_DOWN_LEFT;            
            break;
         
         case PST_CLOSE_PARENTHESIS:
            Depth = PST_PopULongStack(m_DepthStack, &m_DepthStackPointer);
            Label = PST_PopULongStack(m_LabelStack, &m_LabelStackPointer);
            NodeType = PST_PopNodeTypeStack(m_NodeTypeStack, &m_NodeTypeStackPointer);
            PST_MarkThenMoveUpLeft(Depth, Label, NodeType);
            m_CurrentParenthesisCount--;
            m_State = PST_STATE_MOVE_UP_LEFT;            
            break;
         
         default:
            printf("Error: tree string is invalid (4:%c).\n", Ch);
            exit(1);
            break;
      }
   }
}

/*********************************************************************************/

void PST_StateMoveUpLeft (UBYTE Ch, UBYTE *TreeString)
{
   ULONG Depth;
   ULONG Label;
   PST_NODE_TYPE NodeType;
   
   if (!isspace(Ch))
   {
      switch(Ch)
      {
         case PST_QUOTATION_MARK:
            /* Just need to dump this information */
            PST_FindNextDepthAndLabel (TreeString, m_CurrentTreeStringIndex - 1, TRUE, &Depth, &Label, &NodeType);
            m_State = PST_STATE_MOVE_UP_LEFT;            
            break;
            
         case PST_COLON:
            /* Just need to dump this information */
            PST_FindNextDepthAndLabel (TreeString, m_CurrentTreeStringIndex - 1, TRUE, &Depth, &Label, &NodeType);
            m_State = PST_STATE_MOVE_UP_LEFT;            
            break;
         
         case PST_CLOSE_PARENTHESIS:
            Depth = PST_PopULongStack(m_DepthStack, &m_DepthStackPointer);
            Label = PST_PopULongStack(m_LabelStack, &m_LabelStackPointer);
            NodeType = PST_PopNodeTypeStack(m_NodeTypeStack, &m_NodeTypeStackPointer);
            PST_MarkThenMoveUpLeft(Depth, Label, NodeType);
            m_CurrentParenthesisCount--;
            m_State = PST_STATE_MOVE_UP_LEFT;            
            break;
         
         case PST_COMMA:
            Depth = PST_PopULongStack(m_DepthStack, &m_DepthStackPointer);
            Label = PST_PopULongStack(m_LabelStack, &m_LabelStackPointer);
            NodeType = PST_PopNodeTypeStack(m_NodeTypeStack, &m_NodeTypeStackPointer);
            PST_MarkThenMoveUpRight(Depth, Label, NodeType);
            m_State = PST_STATE_MOVE_UP_RIGHT;            
            break;
         
         default:
            printf("Error: tree string is invalid (5:%c).\n", Ch);
            exit(1);
            break;
      }
   }
}

/*********************************************************************************/

ULONG PST_FindIndexOfMatchingToken (UBYTE Token, 
                                    UBYTE *TreeString, 
                                    ULONG CurrentTreeStringIndex)
{
   ULONG TokenCount = 1;
   ULONG TempTreeStringIndex = CurrentTreeStringIndex;
   UBYTE Ch;

   while (TokenCount > 0)
   {
      Ch = TreeString[TempTreeStringIndex++];
      switch (Token)
      {
         case PST_OPEN_PARENTHESIS:
            if (Ch == PST_OPEN_PARENTHESIS)
            {
               TokenCount++;
            }
            else if (Ch == PST_CLOSE_PARENTHESIS)
            {
               TokenCount--;
            }
            break;
            
         case PST_QUOTATION_MARK:
            if (Ch == PST_QUOTATION_MARK)
            {
               TokenCount--;
            }
            break;
            
         default:
            printf("Error: unknown token.\n");
            exit(1);
            break;
      }
   }

   return (TempTreeStringIndex - 1);
}

/*********************************************************************************/

void PST_FindDepthAndLabelAfterMatchingParenthesis (UBYTE *TreeString, 
                                                      ULONG *Depth, 
                                                      ULONG *Label, 
                                                      PST_NODE_TYPE *NodeType)
{
   ULONG TempTreeStringIndex;

   TempTreeStringIndex = PST_FindIndexOfMatchingToken (PST_OPEN_PARENTHESIS, 
                                                         TreeString, 
                                                         m_CurrentTreeStringIndex);
   PST_FindNextDepthAndLabel (TreeString, TempTreeStringIndex, FALSE, Depth, Label, NodeType);
}

/*********************************************************************************/

BOOLEAN PST_FindNextDepthAndLabel (UBYTE *TreeString, 
                                 ULONG CurrentTreeStringIndex, 
                                 BOOLEAN MoveGlobalIndex, 
                                 ULONG *Depth, 
                                 ULONG *Label, 
                                 PST_NODE_TYPE *NodeType)
{
   ULONG TempTreeStringIndex = CurrentTreeStringIndex;
   UBYTE Ch;
   BOOLEAN FoundLabel = FALSE;
   BOOLEAN FoundDepth = FALSE;
   BOOLEAN StartReadingDepth = FALSE;
   BOOLEAN StartReadingLabel = FALSE;

   *Depth = 0;
   *Label = 0;
   *NodeType = PST_NODE_NORMAL;

   while ((FoundLabel == FALSE) && (StartReadingDepth == FALSE))
   {
      Ch = TreeString[TempTreeStringIndex++];
      if (Ch == PST_COLON)
      {
         StartReadingDepth = TRUE;
      }
      else
      {
         if (StartReadingLabel == FALSE)
         {
            if (Ch == PST_QUOTATION_MARK)
            {
               StartReadingLabel = TRUE;
            }
            else if ((Ch >= PST_ZERO) && (Ch <= PST_NINE))
            {
               StartReadingLabel = TRUE;
               (*Label) *= 10;
               (*Label) += (Ch - PST_ZERO);
            }
            else if (Ch == PST_APOSTROPHE)
            {
               (*NodeType) = PST_NODE_REVERSE_COMPLEMENT;
            }
            else if (Ch == PST_ASTERISK)
            {
               (*NodeType) = PST_NODE_SELF_COMPLEMENT;
            }
         }
         else
         {
            if (Ch == PST_QUOTATION_MARK)
            {
               FoundLabel = TRUE;
            }
            else if ((Ch >= PST_ZERO) && (Ch <= PST_NINE))
            {
               (*Label) *= 10;
               (*Label) += (Ch - PST_ZERO);
            }
            else if (Ch == PST_APOSTROPHE)
            {
               (*NodeType) = PST_NODE_REVERSE_COMPLEMENT;
            }
            else if (Ch == PST_ASTERISK)
            {
               (*NodeType) = PST_NODE_SELF_COMPLEMENT;
            }
         }
      }
   }
   
   while (FoundDepth == FALSE)
   {
      Ch = TreeString[TempTreeStringIndex++];
      if (Ch == PST_COLON)
      {
         StartReadingDepth = TRUE;
      }
      else if (StartReadingDepth == TRUE)
      {
         if ((Ch >= PST_ZERO) && (Ch <= PST_NINE))
         {
            (*Depth) *= 10;
            (*Depth) += (Ch - PST_ZERO);
         }
         else
         {
            FoundDepth = TRUE;
         }
      }
   }
   
   if (MoveGlobalIndex == TRUE)
   {
      m_CurrentTreeStringIndex = TempTreeStringIndex - 1;
   }
   
   if (FoundLabel == FALSE)
   {
      (*Label) = PST_NO_LABEL;
   }
   
   return(FoundLabel);
}
