/*****************************************************************/
/*  ELEC 999                                                     */
/*  University of Nebraska-Lincoln                               */
/*  Department of Electrical Engineering                         */
/*  David J. Russell                                             */
/*  9-21-08                                                      */
/*                                                               */
/* 02/14/09 - Added html output type where text is colored to    */
/*             identify the structural pieces. Modified          */
/*             VectorADT to allow longer vectors, version 1.5    */
/*****************************************************************/

#ifdef _VECTORADT
#define VECTORADT_EXTERN
#else
#define VECTORADT_EXTERN extern
#endif

/******************************************************************/
/* Public Module Constants                                        */
/******************************************************************/
#define VECTORADT_MAX_VECTOR_TIMES_THREE  7

/******************************************************************/
/* Public Type Definitions                                        */
/******************************************************************/

typedef struct
{
   /*
      Using static array size in an attempt to speed things up a little.
    */
   ULONG CellsByThree[VECTORADT_MAX_VECTOR_TIMES_THREE];
   UBYTE Length;
} VECTORADT_VECTOR_TYPE;

typedef enum
{
   VECTORADT_AGTB,
   VECTORADT_AEQB,
   VECTORADT_ALTB
} VECTORADT_EQUALITY_TYPE;

/*********************************************************************************/
/* Public Module Variables */
/*********************************************************************************/

/*********************************************************************************/
/* Public Module Function Prototypes */
/*********************************************************************************/
void VECTORADT_InitializeVector(VECTORADT_VECTOR_TYPE *Vector);
void VECTORADT_SetElement(VECTORADT_VECTOR_TYPE *Vector, ULONG Element, UBYTE CellIndex);
void VECTORADT_VectorSum(VECTORADT_VECTOR_TYPE *A, 
                           VECTORADT_VECTOR_TYPE *B, 
                           VECTORADT_VECTOR_TYPE *C);
void VECTORADT_VectorAbsDifference(VECTORADT_VECTOR_TYPE *A, 
                                    VECTORADT_VECTOR_TYPE *B, 
                                    VECTORADT_VECTOR_TYPE *C);
VECTORADT_EQUALITY_TYPE VECTORADT_VectorCompare(VECTORADT_VECTOR_TYPE *A, 
                                                VECTORADT_VECTOR_TYPE *B);
void VECTORADT_PrintVector(VECTORADT_VECTOR_TYPE *Vector, FILE *FileStream);
